from __future__ import annotations

from typing import TYPE_CHECKING, Dict

import nbformat

from great_expectations.render.renderer.notebook_renderer import BaseNotebookRenderer

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )


class CheckpointNewNotebookRenderer(BaseNotebookRenderer):
    def __init__(self, context: AbstractDataContext, checkpoint_name: str) -> None:
        super().__init__(context=context)
        self.context = context
        self.checkpoint_name = checkpoint_name

    def _find_datasource_with_asset(self) -> Dict[str, str]:
        """
        Find a Datasource with a configured Asset.

        Useful to pre-populate a working sample Checkpoint for notebook users.
        """
        datasource_candidate = None
        for datasource_name, datasource in self.context.datasources.items():
            for (
                data_connector_name,
                data_connector,
            ) in datasource.data_connectors.items():
                if len(data_connector.get_available_data_asset_names()) > 0:
                    datasource_candidate = {
                        "datasource_name": datasource_name,
                        "data_connector_name": data_connector_name,
                        "asset_name": list(
                            data_connector.get_available_data_asset_names()
                        )[0],
                    }
                    break
        return datasource_candidate

    def _add_header(self) -> None:
        self.add_markdown_cell(
            f"""# Create Your Checkpoint
Use this notebook to configure a new Checkpoint and add it to your project:

**Checkpoint Name**: `{self.checkpoint_name}`"""
        )

    def _add_imports(self) -> None:
        self.add_code_cell(
            """from ruamel.yaml import YAML
import great_expectations as gx
from pprint import pprint

yaml = YAML()
context = gx.get_context()
""",
            lint=True,
        )

    def _add_optional_customize_your_config(self) -> None:
        self.add_markdown_cell(
            """# Customize Your Configuration
The following cells show examples for listing your current configuration. You can replace values in the sample configuration with these values to customize your Checkpoint."""
        )
        self.add_code_cell(
            """# Run this cell to print out the names of your Datasources, Data Connectors and Data Assets
pprint(context.get_available_data_asset_names())""",
            lint=True,
        )
        self.add_code_cell("context.list_expectation_suite_names()")

    def _add_sample_checkpoint_config(self) -> None:
        self.add_markdown_cell(
            """# Create a Checkpoint Configuration

**If you are new to Great Expectations or the Checkpoint feature**, you should start with SimpleCheckpoint because it includes default configurations like a default list of post validation actions.

In the cell below we have created a sample Checkpoint configuration using **your configuration** and **SimpleCheckpoint** to run a single validation of a single Expectation Suite against a single Batch of data.

To keep it simple, we are just choosing the first available instance of each of the following items you have configured in your Data Context:
* Datasource
* DataConnector
* DataAsset
* Partition
* Expectation Suite

Of course this is purely an example, you may edit this to your heart's content.

**My configuration is not so simple - are there more advanced options?**

Glad you asked! Checkpoints are very versatile. For example, you can validate many Batches in a single Checkpoint, validate Batches against different Expectation Suites or against many Expectation Suites, control the specific post-validation actions based on Expectation Suite / Batch / results of validation among other features. Check out our documentation on Checkpoints for more details and for instructions on how to implement other more advanced features including using the **Checkpoint** class:
- https://docs.greatexpectations.io/docs/reference/checkpoints_and_actions
- https://docs.greatexpectations.io/docs/guides/validation/checkpoints/how_to_create_a_new_checkpoint
- https://docs.greatexpectations.io/docs/guides/validation/checkpoints/how_to_configure_a_new_checkpoint_using_test_yaml_config"""
        )
        try:
            first_datasource_with_asset = self._find_datasource_with_asset()
            first_datasource_name = first_datasource_with_asset["datasource_name"]
            first_data_connector_name = first_datasource_with_asset[
                "data_connector_name"
            ]
            first_asset_name = first_datasource_with_asset["asset_name"]

            first_expectation_suite = self.context.list_expectation_suites()[0]
            first_expectation_suite_name = (
                first_expectation_suite.expectation_suite_name
            )
            sample_yaml_str = f'my_checkpoint_name = "{self.checkpoint_name}" # This was populated from your CLI command.\n\n'
            sample_yaml_str += 'yaml_config = f"""'
            sample_yaml_str += "\nname: {my_checkpoint_name}"
            sample_yaml_str += f"""
config_version: 1.0
class_name: SimpleCheckpoint
run_name_template: "%Y%m%d-%H%M%S-my-run-name-template"
validations:
  - batch_request:
      datasource_name: {first_datasource_name}
      data_connector_name: {first_data_connector_name}
      data_asset_name: {first_asset_name}
      data_connector_query:
        index: -1
    expectation_suite_name: {first_expectation_suite_name}
"""
            sample_yaml_str += '"""'
            sample_yaml_str += "\nprint(yaml_config)"

            self.add_code_cell(
                sample_yaml_str,
                lint=True,
            )
        except Exception:
            # For any error
            self.add_markdown_cell(
                "Sorry, we were unable to create a sample configuration. Perhaps you don't have a Datasource or Expectation Suite configured."
            )

    def _add_test_and_save_your_checkpoint_configuration(self) -> None:
        self.add_markdown_cell(
            """# Test Your Checkpoint Configuration
Here we will test your Checkpoint configuration to make sure it is valid.

This `test_yaml_config()` function is meant to enable fast dev loops. If your configuration is correct, this cell will show a message that you successfully instantiated a Checkpoint. You can continually edit your Checkpoint config yaml and re-run the cell to check until the new config is valid.

If you instead wish to use python instead of yaml to configure your Checkpoint, you can use `context.add_checkpoint()` and specify all the required parameters."""
        )
        self.add_code_cell(
            """my_checkpoint = context.test_yaml_config(yaml_config=yaml_config)""",
            lint=True,
        )

    def _add_review_checkpoint(self) -> None:
        self.add_markdown_cell(
            """# Review Your Checkpoint

You can run the following cell to print out the full yaml configuration. For example, if you used **SimpleCheckpoint**  this will show you the default action list."""
        )
        self.add_code_cell('print(my_checkpoint.get_config(mode="yaml"))', lint=True)

    def _add_add_checkpoint(self) -> None:
        self.add_markdown_cell(
            """# Add Your Checkpoint

Run the following cell to save this Checkpoint to your Checkpoint Store."""
        )
        self.add_code_cell(
            "context.add_checkpoint(**yaml.load(yaml_config))",
            lint=True,
        )

    def _add_optional_run_checkpoint(self) -> None:
        self.add_markdown_cell(
            """# Run Your Checkpoint & Open Data Docs(Optional)

You may wish to run the Checkpoint now and review its output in Data Docs. If so uncomment and run the following cell."""
        )
        self.add_code_cell(
            "# context.run_checkpoint(checkpoint_name=my_checkpoint_name)\n# context.open_data_docs()",
            lint=True,
        )

    def render(self) -> nbformat.NotebookNode:
        self._notebook: nbformat.NotebookNode = nbformat.v4.new_notebook()
        self._add_header()
        self._add_imports()
        self._add_sample_checkpoint_config()
        self._add_optional_customize_your_config()
        self._add_test_and_save_your_checkpoint_configuration()
        self._add_review_checkpoint()
        self._add_add_checkpoint()
        self._add_optional_run_checkpoint()

        return self._notebook

    def render_to_disk(
        self,
        notebook_file_path: str,
    ) -> None:
        self.render()
        self.write_notebook_to_disk(self._notebook, notebook_file_path)



from .query_column import QueryColumn
from .query_column_pair import QueryColumnPair
from .query_multiple_columns import QueryMultipleColumns
from .query_table import QueryTable
from .query_template_values import QueryTemplateValues



from ruamel import yaml

import great_expectations as gx
from great_expectations.checkpoint import SimpleCheckpoint
from great_expectations.core.batch import BatchRequest
from great_expectations.profile.user_configurable_profiler import (
    UserConfigurableProfiler,
)
from great_expectations.validator.validator import Validator

from great_expectations.core.usage_statistics.anonymizers.types.base import (  # isort:skip
    GETTING_STARTED_DATASOURCE_NAME,
    GETTING_STARTED_EXPECTATION_SUITE_NAME,
    GETTING_STARTED_CHECKPOINT_NAME,
)

context = gx.get_context()
# NOTE: The following assertion is only for testing and can be ignored by users.
assert context

# First configure a new Datasource and add to DataContext

# <snippet name="tests/integration/docusaurus/tutorials/getting-started/getting_started.py datasource_yaml">
datasource_yaml = f"""
name: getting_started_datasource
class_name: Datasource
execution_engine:
  class_name: PandasExecutionEngine
data_connectors:
    default_inferred_data_connector_name:
        class_name: InferredAssetFilesystemDataConnector
        base_directory: ../data/
        default_regex:
          group_names:
            - data_asset_name
          pattern: (.*)
    default_runtime_data_connector_name:
        class_name: RuntimeDataConnector
        assets:
            my_runtime_asset_name:
              batch_identifiers:
                - runtime_batch_identifier_name
"""
# </snippet>

# Note : this override is for internal GX purposes, and is intended to helps us better understand how the
# Getting Started Guide is being used. It can be ignored by users.
datasource_yaml = datasource_yaml.replace(
    "getting_started_datasource", GETTING_STARTED_DATASOURCE_NAME
)

context.test_yaml_config(datasource_yaml)
context.add_datasource(**yaml.load(datasource_yaml))

# Get Validator by creating ExpectationSuite and passing in BatchRequest
batch_request = BatchRequest(
    datasource_name="getting_started_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="yellow_tripdata_sample_2019-01.csv",
    limit=1000,
)

# Note : this override is for internal GX purposes, and is intended to helps us better understand how the
# Getting Started Guide is being used. It can be ignored by users.
batch_request = BatchRequest(
    datasource_name=GETTING_STARTED_DATASOURCE_NAME,
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="yellow_tripdata_sample_2019-01.csv",
    limit=1000,
)

expectation_suite_name = "getting_started_expectation_suite_taxi.demo"

# Note : this override is for internal GX purposes, and is intended to helps us better understand how the
# Getting Started Guide is being used. It can be ignored by users
expectation_suite_name = GETTING_STARTED_EXPECTATION_SUITE_NAME

context.add_or_update_expectation_suite(expectation_suite_name=expectation_suite_name)

validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name=expectation_suite_name
)

# NOTE: The following assertion is only for testing and can be ignored by users.
assert isinstance(validator, Validator)

# Profile the data with the UserConfigurableProfiler and save resulting ExpectationSuite
# <snippet name="tests/integration/docusaurus/tutorials/getting-started/getting_started.py exclude_column_names no comment">
exclude_column_names = [
    "vendor_id",
    "pickup_datetime",
    "dropoff_datetime",
    # "passenger_count",
    "trip_distance",
    "rate_code_id",
    "store_and_fwd_flag",
    "pickup_location_id",
    "dropoff_location_id",
    "payment_type",
    "fare_amount",
    "extra",
    "mta_tax",
    "tip_amount",
    "tolls_amount",
    "improvement_surcharge",
    "total_amount",
    "congestion_surcharge",
]
# </snippet>

profiler = UserConfigurableProfiler(
    profile_dataset=validator,
    excluded_expectations=None,
    ignored_columns=exclude_column_names,
    not_null_only=False,
    primary_or_compound_key=None,
    semantic_types_dict=None,
    table_expectations_only=False,
    value_set_threshold="MANY",
)
suite = profiler.build_suite()
validator.expectation_suite = suite
validator.save_expectation_suite(discard_failed_expectations=False)

# Create first checkpoint on yellow_tripdata_sample_2019-01.csv
my_checkpoint_config = f"""
name: getting_started_checkpoint
config_version: 1.0
class_name: SimpleCheckpoint
run_name_template: "%Y%m%d-%H%M%S-my-run-name-template"
validations:
  - batch_request:
      datasource_name: getting_started_datasource
      data_connector_name: default_inferred_data_connector_name
      data_asset_name: yellow_tripdata_sample_2019-01.csv
      data_connector_query:
        index: -1
    expectation_suite_name: getting_started_expectation_suite_taxi.demo
"""
# Note : these overrides are for internal GX purposes, and are intended to helps us better understand how the
# Getting Started Guide is being used. It can be ignored by users
my_checkpoint_config = my_checkpoint_config.replace(
    "getting_started_checkpoint", GETTING_STARTED_CHECKPOINT_NAME
)
yaml_config = my_checkpoint_config.replace(
    "getting_started_datasource", GETTING_STARTED_DATASOURCE_NAME
)
my_checkpoint_config = my_checkpoint_config.replace(
    "getting_started_expectation_suite_taxi.demo",
    GETTING_STARTED_EXPECTATION_SUITE_NAME,
)


my_checkpoint_config = yaml.load(my_checkpoint_config)

# NOTE: The following code (up to and including the assert) is only for testing and can be ignored by users.
# In the current test, site_names are set to None because we do not want to update and build data_docs
# If you would like to build data_docs then either remove `site_names=None` or pass in a list of site_names you would like to build the docs on.
checkpoint = SimpleCheckpoint(
    **my_checkpoint_config, data_context=context, site_names=None
)
checkpoint_result = checkpoint.run(site_names=None)
assert checkpoint_result.run_results


# Create second checkpoint on yellow_tripdata_sample_2019-02.csv
# <snippet name="tests/integration/docusaurus/tutorials/getting-started/getting_started.py checkpoint_yaml_config">
yaml_config = f"""
name: getting_started_checkpoint
config_version: 1.0
class_name: SimpleCheckpoint
run_name_template: "%Y%m%d-%H%M%S-my-run-name-template"
validations:
  - batch_request:
      datasource_name: getting_started_datasource
      data_connector_name: default_inferred_data_connector_name
      data_asset_name: yellow_tripdata_sample_2019-02.csv
      data_connector_query:
        index: -1
    expectation_suite_name: getting_started_expectation_suite_taxi.demo
"""
# </snippet>
# Note : this override is for internal GX purposes, and is intended to helps us better understand how the
# Getting Started Guide is being used. It can be ignored by users
yaml_config = yaml_config.replace(
    "getting_started_checkpoint", GETTING_STARTED_CHECKPOINT_NAME
)
yaml_config = yaml_config.replace(
    "getting_started_datasource", GETTING_STARTED_DATASOURCE_NAME
)
yaml_config = yaml_config.replace(
    "getting_started_expectation_suite_taxi.demo",
    GETTING_STARTED_EXPECTATION_SUITE_NAME,
)

my_new_checkpoint_config = yaml.load(yaml_config)

# NOTE: The following code (up to and including the assert) is only for testing and can be ignored by users.
# In the current test, site_names are set to None because we do not want to update and build data_docs
# If you would like to build data_docs then either remove `site_names=None` or pass in a list of site_names you would like to build the docs on.
new_checkpoint = SimpleCheckpoint(
    **my_new_checkpoint_config, data_context=context, site_names=None
)
new_checkpoint_result = new_checkpoint.run(site_names=None)
assert new_checkpoint_result.run_results



from __future__ import annotations

import copy
import datetime
import itertools
import logging
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Set,
    Union,
)

import numpy as np

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.domain import Domain  # noqa: TCH001
from great_expectations.rule_based_profiler.config import (
    ParameterBuilderConfig,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.estimators.bootstrap_numeric_range_estimator import (
    BootstrapNumericRangeEstimator,
)
from great_expectations.rule_based_profiler.estimators.exact_numeric_range_estimator import (
    ExactNumericRangeEstimator,
)
from great_expectations.rule_based_profiler.estimators.kde_numeric_range_estimator import (
    KdeNumericRangeEstimator,
)
from great_expectations.rule_based_profiler.estimators.numeric_range_estimation_result import (
    NUM_HISTOGRAM_BINS,
    NumericRangeEstimationResult,
)
from great_expectations.rule_based_profiler.estimators.numeric_range_estimator import (
    NumericRangeEstimator,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.estimators.quantiles_numeric_range_estimator import (
    QuantilesNumericRangeEstimator,
)
from great_expectations.rule_based_profiler.helpers.util import (
    NP_EPSILON,
    build_numeric_range_estimation_result,
    datetime_semantic_domain_type,
    get_parameter_value_and_validate_return_type,
    integer_semantic_domain_type,
)
from great_expectations.rule_based_profiler.metric_computation_result import (
    MetricValues,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.parameter_builder import (
    MetricMultiBatchParameterBuilder,
)
from great_expectations.rule_based_profiler.parameter_container import (
    FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY,
    FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY,
    RAW_PARAMETER_KEY,
    ParameterContainer,
    ParameterNode,
)
from great_expectations.types.attributes import Attributes
from great_expectations.util import (
    convert_ndarray_decimal_to_float_dtype,
    does_ndarray_contain_decimal_dtype,
    is_ndarray_datetime_dtype,
    is_numeric,
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

MAX_DECIMALS: int = 9


class NumericMetricRangeMultiBatchParameterBuilder(MetricMultiBatchParameterBuilder):
    """
    A Multi-Batch implementation for obtaining the range estimation bounds for a resolved (evaluated) numeric metric,
    using domain_kwargs, value_kwargs, metric_name, and false_positive_rate (tolerance) as arguments.

    This Multi-Batch ParameterBuilder is general in the sense that any metric that computes numbers can be accommodated.
    On the other hand, it is specific in the sense that the parameter names will always have the semantics of numeric
    ranges, which will incorporate the requirements, imposed by the configured false_positive_rate tolerances.

    The implementation supports four methods of estimating parameter values from data:
    * quantiles -- assumes that metric values, computed on batch data, are normally distributed and computes the mean
      and the standard error using the queried batches as the single sample of the distribution.
    * exact -- uses the minimum and maximum observations for range boundaries.
    * bootstrap -- a statistical resampling technique (see "https://en.wikipedia.org/wiki/Bootstrapping_(statistics)").
    * kde -- a statistical technique that fits a gaussian to the distribution and resamples from it.
    """

    RECOGNIZED_SAMPLING_METHOD_NAMES: set = {
        "bootstrap",
        "exact",
        "kde",
        "quantiles",
    }

    RECOGNIZED_TRUNCATE_DISTRIBUTION_KEYS: set = {
        "lower_bound",
        "upper_bound",
    }

    METRIC_NAMES_EXEMPT_FROM_VALUE_ROUNDING: set = {
        "column.unique_proportion",
    }

    exclude_field_names: ClassVar[
        Set[str]
    ] = MetricMultiBatchParameterBuilder.exclude_field_names | {
        "single_batch_mode",
    }

    def __init__(
        self,
        name: str,
        metric_name: Optional[str] = None,
        metric_multi_batch_parameter_builder_name: Optional[str] = None,
        metric_domain_kwargs: Optional[Union[str, dict]] = None,
        metric_value_kwargs: Optional[Union[str, dict]] = None,
        enforce_numeric_metric: Union[str, bool] = True,
        replace_nan_with_zero: Union[str, bool] = True,
        reduce_scalar_metric: Union[str, bool] = True,
        false_positive_rate: Optional[Union[str, float]] = None,
        estimator: str = "bootstrap",
        n_resamples: Optional[Union[str, int]] = None,
        random_seed: Optional[Union[str, int]] = None,
        quantile_statistic_interpolation_method: str = "auto",
        quantile_bias_correction: Union[str, bool] = False,
        quantile_bias_std_error_ratio_threshold: Optional[Union[str, float]] = None,
        bw_method: Optional[Union[str, float, Callable]] = None,
        include_estimator_samples_histogram_in_details: Union[str, bool] = False,
        truncate_values: Optional[
            Union[str, Dict[str, Union[Optional[int], Optional[float]]]]
        ] = None,
        round_decimals: Optional[Union[str, int]] = None,
        evaluation_parameter_builder_configs: Optional[
            List[ParameterBuilderConfig]
        ] = None,
        data_context: Optional[AbstractDataContext] = None,
    ) -> None:
        """
        Args:
            name: the name of this parameter -- this is user-specified parameter name (from configuration); it is not
                the fully-qualified parameter name; a fully-qualified parameter name must start with "$parameter."
                and may contain one or more subsequent parts (e.g., "$parameter.<my_param_from_config>.<metric_name>").
            metric_name: the name of a metric used in MetricConfiguration (must be a supported and registered metric)
            metric_multi_batch_parameter_builder_name: name of parameter that computes "metric_name" (for every Batch).
            metric_domain_kwargs: used in MetricConfiguration
            metric_value_kwargs: used in MetricConfiguration
            enforce_numeric_metric: used in MetricConfiguration to insure that metric computations return numeric values
            replace_nan_with_zero: if False, then if the computed metric gives NaN, then exception is raised; otherwise,
                if True (default), then if the computed metric gives NaN, then it is converted to the 0.0 (float) value.
            reduce_scalar_metric: if True (default), then reduces computation of 1-dimensional metric to scalar value.
            false_positive_rate: user-configured fraction between 0 and 1 expressing desired false positive rate for
                identifying unexpected values as judged by the upper- and lower- quantiles of the observed metric data.
            estimator: choice of the estimation algorithm: "quantiles", "bootstrap", "exact"
                (deterministic, incorporating entire observed value range), or "kde" (kernel density estimation).
            n_resamples: Applicable only for the "bootstrap" and "kde" sampling methods -- if omitted (default), then
                9999 is used (default in
                "https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.bootstrap.html").
            random_seed: Applicable only for the "bootstrap" and "kde" sampling methods -- if omitted (default), then
                uses "np.random.choice"; otherwise, utilizes "np.random.Generator(np.random.PCG64(random_seed))".
            quantile_statistic_interpolation_method: Supplies value of (interpolation) "method" to "np.quantile()"
                statistic, used for confidence intervals.
            quantile_bias_correction: Applicable only for the "bootstrap" sampling method -- if omitted (default), then
                False is used (bias correction is disabled) and quantile_bias_std_error_ratio_threshold will have no
                effect.
            quantile_bias_std_error_ratio_threshold: Applicable only for the "bootstrap" sampling method -- if omitted
                (default), then 0.25 is used (as minimum ratio of bias to standard error for applying bias correction).
            bw_method: Applicable only for the "kde" sampling method -- if omitted (default), then "scott" is used.
                Possible values for the estimator bandwidth method are described at:
                https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html
            include_estimator_samples_histogram_in_details: Applicable only for the "bootstrap" and "kde" sampling
                methods -- if True, then add 10-bin histogram of metric samples to "details"; otherwise, omit (default).
            truncate_values: user-configured directive for whether or not to allow the computed parameter values
                (i.e., lower_bound, upper_bound) to take on values outside the specified bounds when packaged on output.
            round_decimals: user-configured non-negative integer indicating the number of decimals of the
                rounding precision of the computed parameter values (i.e., min_value, max_value) prior to packaging them
                on output.  If omitted, then no rounding is performed, unless the computed value is already an integer.
            evaluation_parameter_builder_configs: ParameterBuilder configurations, executing and making whose respective
                ParameterBuilder objects' outputs available (as fully-qualified parameter names) is pre-requisite.
                These "ParameterBuilder" configurations help build parameters needed for this "ParameterBuilder".
            data_context: AbstractDataContext associated with this ParameterBuilder
        """
        super().__init__(
            name=name,
            metric_name=metric_name,
            metric_domain_kwargs=metric_domain_kwargs,
            metric_value_kwargs=metric_value_kwargs,
            single_batch_mode=False,
            enforce_numeric_metric=enforce_numeric_metric,
            replace_nan_with_zero=replace_nan_with_zero,
            reduce_scalar_metric=reduce_scalar_metric,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
            data_context=data_context,
        )

        self._metric_multi_batch_parameter_builder_name = (
            metric_multi_batch_parameter_builder_name
        )

        if false_positive_rate is None:
            false_positive_rate = 5.0e-2

        self._false_positive_rate = false_positive_rate

        self._estimator = estimator

        self._n_resamples = n_resamples

        self._random_seed = random_seed

        self._quantile_statistic_interpolation_method = (
            quantile_statistic_interpolation_method
        )

        self._quantile_bias_correction = quantile_bias_correction

        self._quantile_bias_std_error_ratio_threshold = (
            quantile_bias_std_error_ratio_threshold
        )

        self._bw_method = bw_method

        self._include_estimator_samples_histogram_in_details = (
            include_estimator_samples_histogram_in_details
        )

        self._round_decimals = round_decimals

        if not truncate_values:
            truncate_values = {
                "lower_bound": None,
                "upper_bound": None,
            }
        else:
            if not isinstance(truncate_values, str):
                truncate_values_keys: set = set(truncate_values.keys())
                if (
                    not truncate_values_keys
                    <= NumericMetricRangeMultiBatchParameterBuilder.RECOGNIZED_TRUNCATE_DISTRIBUTION_KEYS
                ):
                    raise gx_exceptions.ProfilerExecutionError(
                        message=f"""Unrecognized truncate_values key(s) in {self.__class__.__name__}:
"{str(truncate_values_keys - NumericMetricRangeMultiBatchParameterBuilder.RECOGNIZED_TRUNCATE_DISTRIBUTION_KEYS)}" \
detected.
"""
                    )

        self._truncate_values = truncate_values

    @property
    def metric_multi_batch_parameter_builder_name(self) -> str:
        return self._metric_multi_batch_parameter_builder_name

    @property
    def false_positive_rate(self) -> Union[str, float]:
        return self._false_positive_rate

    @property
    def estimator(self) -> str:
        return self._estimator

    @property
    def n_resamples(self) -> Optional[Union[str, int]]:
        return self._n_resamples

    @property
    def random_seed(self) -> Optional[Union[str, int]]:
        return self._random_seed

    @property
    def quantile_statistic_interpolation_method(self) -> str:
        return self._quantile_statistic_interpolation_method

    @property
    def quantile_bias_correction(self) -> Union[str, bool]:
        return self._quantile_bias_correction

    @property
    def quantile_bias_std_error_ratio_threshold(self) -> Optional[Union[str, float]]:
        return self._quantile_bias_std_error_ratio_threshold

    @property
    def bw_method(self) -> Optional[Union[str, float, Callable]]:
        return self._bw_method

    @property
    def include_estimator_samples_histogram_in_details(self) -> Union[str, bool]:
        return self._include_estimator_samples_histogram_in_details

    @property
    def truncate_values(
        self,
    ) -> Optional[Union[str, Dict[str, Union[Optional[int], Optional[float]]]]]:
        return self._truncate_values

    @property
    def round_decimals(self) -> Optional[Union[str, int]]:
        return self._round_decimals

    def _build_parameters(
        self,
        domain: Domain,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> Attributes:
        """
        Builds ParameterContainer object that holds ParameterNode objects with attribute name-value pairs and details.

        Returns:
            Attributes object, containing computed parameter values and parameter computation details metadata.

         The algorithm operates according to the following steps:
         1. Obtain batch IDs of interest using AbstractDataContext and BatchRequest (unless passed explicitly as argument).
         2. Set up metric_domain_kwargs and metric_value_kwargs (using configuration and/or variables and parameters).
         3. Instantiate the Validator object corresponding to BatchRequest (with a temporary expectation_suite_name) in
            order to have access to all Batch objects, on each of which the specified metric_name will be computed.
         4. Perform metric computations and obtain the result in the array-like form (one metric value per each Batch).
         5. Using the configured directives and heuristics, determine whether or not the ranges should be clipped.
         6. Using the configured directives and heuristics, determine if return values should be rounded to an integer.
         7. Convert the multi-dimensional metric computation results to a numpy array (for further computations).
         8. Compute [low, high] for the desired metric using the chosen estimator method.
         9. Return [low, high] for the desired metric as estimated by the specified sampling method.
        10. Set up the arguments and call build_parameter_container() to store the parameter as part of "rule state".
        """
        parameter_reference: str
        if self.metric_multi_batch_parameter_builder_name:
            # Obtain metric_multi_batch_parameter_builder_name from "rule state" (i.e., variables and parameters); from instance variable otherwise.
            metric_multi_batch_parameter_builder_name: str = (
                get_parameter_value_and_validate_return_type(
                    domain=domain,
                    parameter_reference=self.metric_multi_batch_parameter_builder_name,
                    expected_return_type=str,
                    variables=variables,
                    parameters=parameters,
                )
            )
            parameter_reference = (
                f"{RAW_PARAMETER_KEY}{metric_multi_batch_parameter_builder_name}"
            )
        else:
            # Compute metric value for each Batch object.
            super().build_parameters(
                domain=domain,
                variables=variables,
                parameters=parameters,
                parameter_computation_impl=super()._build_parameters,
                runtime_configuration=runtime_configuration,
            )
            parameter_reference = self.raw_fully_qualified_parameter_name

        # Retrieve metric values for all Batch objects.
        parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
            domain=domain,
            parameter_reference=parameter_reference,
            expected_return_type=None,
            variables=variables,
            parameters=parameters,
        )
        metric_values: MetricValues = parameter_node[
            FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY
        ]

        round_decimals: int
        if (
            self.metric_name
            not in NumericMetricRangeMultiBatchParameterBuilder.METRIC_NAMES_EXEMPT_FROM_VALUE_ROUNDING
            and integer_semantic_domain_type(domain=domain)
        ):
            round_decimals = 0
        else:
            round_decimals = self._get_round_decimals_using_heuristics(
                metric_values=metric_values,
                domain=domain,
                variables=variables,
                parameters=parameters,
            )

        numeric_range_estimator: NumericRangeEstimator = (
            self._build_numeric_range_estimator(
                round_decimals=round_decimals,
                domain=domain,
                variables=variables,
                parameters=parameters,
            )
        )
        numeric_range_estimation_result: NumericRangeEstimationResult = (
            self._estimate_metric_value_range(
                metric_values=metric_values,
                numeric_range_estimator=numeric_range_estimator,
                round_decimals=round_decimals,
                domain=domain,
                variables=variables,
                parameters=parameters,
            )
        )

        value_range: np.ndarray = numeric_range_estimation_result.value_range
        details: Dict[str, Any] = copy.deepcopy(
            parameter_node[FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY]
        )

        # Obtain include_estimator_samples_histogram_in_details from "rule state" (i.e., variables and parameters); from instance variable otherwise.
        include_estimator_samples_histogram_in_details: bool = (
            get_parameter_value_and_validate_return_type(
                domain=domain,
                parameter_reference=self.include_estimator_samples_histogram_in_details,
                expected_return_type=bool,
                variables=variables,
                parameters=parameters,
            )
        )

        if include_estimator_samples_histogram_in_details:
            details[
                "estimation_histogram"
            ] = numeric_range_estimation_result.estimation_histogram

        return Attributes(
            {
                FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY: value_range,
                FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY: details,
            }
        )

    def _build_numeric_range_estimator(
        self,
        round_decimals: int,
        domain: Domain,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
    ) -> NumericRangeEstimator:
        """
        Determines "estimator" name and returns appropriate configured "NumericRangeEstimator" subclass instance.
        """
        # Obtain estimator directive from "rule state" (i.e., variables and parameters); from instance variable otherwise.
        estimator: str = get_parameter_value_and_validate_return_type(
            domain=domain,
            parameter_reference=self.estimator,
            expected_return_type=str,
            variables=variables,
            parameters=parameters,
        )
        if (
            estimator
            not in NumericMetricRangeMultiBatchParameterBuilder.RECOGNIZED_SAMPLING_METHOD_NAMES
        ):
            raise gx_exceptions.ProfilerExecutionError(
                message=f"""The directive "estimator" for {self.__class__.__name__} can be only one of
{NumericMetricRangeMultiBatchParameterBuilder.RECOGNIZED_SAMPLING_METHOD_NAMES} ("{estimator}" was detected).
"""
            )

        if estimator == "exact":
            return ExactNumericRangeEstimator()

        if estimator == "quantiles":
            return QuantilesNumericRangeEstimator(
                configuration=Attributes(
                    {
                        "false_positive_rate": self.false_positive_rate,
                        "round_decimals": round_decimals,
                        "quantile_statistic_interpolation_method": self.quantile_statistic_interpolation_method,
                    }
                )
            )

        # Since complex numerical calculations do not support DateTime/TimeStamp data types, use "quantiles" estimator.
        if datetime_semantic_domain_type(domain=domain):
            logger.info(
                f'Estimator "{estimator}" does not support DateTime/TimeStamp data types (downgrading to "quantiles").'
            )
            return QuantilesNumericRangeEstimator(
                configuration=Attributes(
                    {
                        "false_positive_rate": self.false_positive_rate,
                        "round_decimals": round_decimals,
                        "quantile_statistic_interpolation_method": self.quantile_statistic_interpolation_method,
                    }
                )
            )

        if estimator == "bootstrap":
            return BootstrapNumericRangeEstimator(
                configuration=Attributes(
                    {
                        "false_positive_rate": self.false_positive_rate,
                        "round_decimals": round_decimals,
                        "n_resamples": self.n_resamples,
                        "random_seed": self.random_seed,
                        "quantile_statistic_interpolation_method": self.quantile_statistic_interpolation_method,
                        "quantile_bias_correction": self.quantile_bias_correction,
                        "quantile_bias_std_error_ratio_threshold": self.quantile_bias_std_error_ratio_threshold,
                    }
                )
            )

        if estimator == "kde":
            return KdeNumericRangeEstimator(
                configuration=Attributes(
                    {
                        "false_positive_rate": self.false_positive_rate,
                        "round_decimals": round_decimals,
                        "n_resamples": self.n_resamples,
                        "random_seed": self.random_seed,
                        "quantile_statistic_interpolation_method": self.quantile_statistic_interpolation_method,
                        "bw_method": self.bw_method,
                    }
                )
            )

        return ExactNumericRangeEstimator()

    def _estimate_metric_value_range(
        self,
        metric_values: np.ndarray,
        numeric_range_estimator: NumericRangeEstimator,
        round_decimals: int,
        domain: Optional[Domain] = None,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
    ) -> NumericRangeEstimationResult:
        """
        This method accepts "NumericRangeEstimator" and data samples in format "N x R^m", where "N" (most significant
        dimension) is the number of measurements (e.g., one per Batch of data), while "R^m" is the multi-dimensional
        metric, whose values are being estimated.  Thus, for each element in the "R^m" hypercube, an "N"-dimensional
        vector of sample measurements is constructed and given to the estimator to apply its specific algorithm for
        computing the range of values in this vector.  Estimator algorithms differ based on their use of data samples.
        """
        truncate_values: Dict[str, Number] = self._get_truncate_values_using_heuristics(
            metric_values=metric_values,
            domain=domain,
            variables=variables,
            parameters=parameters,
        )
        lower_bound: Optional[float] = truncate_values.get("lower_bound")
        upper_bound: Optional[float] = truncate_values.get("upper_bound")

        min_value: Number
        max_value: Number

        # Outer-most dimension is data samples (e.g., one per Batch); the rest are dimensions of the actual metric.
        metric_value_shape: tuple = metric_values.shape[1:]

        # Generate all permutations of indexes for accessing every element of the multi-dimensional metric.
        metric_value_shape_idx: int
        axes: List[np.ndarray] = [
            np.indices(dimensions=(metric_value_shape_idx,))[0]
            for metric_value_shape_idx in metric_value_shape
        ]
        metric_value_indices: List[tuple] = list(itertools.product(*tuple(axes)))

        # Generate all permutations of indexes for accessing estimates of every element of the multi-dimensional metric.
        # Prefixing multi-dimensional index with "(slice(None, None, None),)" is equivalent to "[:,]" access.
        metric_value_idx: tuple
        metric_value_vector_indices: List[tuple] = [
            (slice(None, None, None),) + metric_value_idx
            for metric_value_idx in metric_value_indices
        ]

        # Initialize value range estimate for multi-dimensional metric to all trivial values (to be updated in situ).
        # Since range includes min and max values, value range estimate contains 2-element least-significant dimension.
        metric_value_range_shape: tuple = metric_value_shape + (2,)
        # Initialize observed_values for multi-dimensional metric to all trivial values (to be updated in situ).
        # Return values of "numpy.histogram()", histogram and bin edges, are packaged in least-significant dimensions.
        estimation_histogram_shape: tuple = metric_value_shape + (
            2,
            NUM_HISTOGRAM_BINS + 1,
        )

        datetime_detected: bool = datetime_semantic_domain_type(
            domain=domain
        ) or self._is_metric_values_ndarray_datetime_dtype(
            metric_values=metric_values,
            metric_value_vector_indices=metric_value_vector_indices,
        )

        metric_value_range: np.ndarray
        estimation_histogram: np.ndarray
        if datetime_detected:
            metric_value_range = np.full(
                shape=metric_value_range_shape, fill_value=datetime.datetime.min
            )
            estimation_histogram = np.full(
                shape=estimation_histogram_shape, fill_value=datetime.datetime.min
            )
        else:
            if self._is_metric_values_ndarray_decimal_dtype(
                metric_values=metric_values,
                metric_value_vector_indices=metric_value_vector_indices,
            ):
                metric_values = convert_ndarray_decimal_to_float_dtype(
                    data=metric_values
                )

            metric_value_range = np.zeros(shape=metric_value_range_shape)
            estimation_histogram = np.empty(shape=estimation_histogram_shape)

        # Traverse indices of sample vectors corresponding to every element of multi-dimensional metric.
        metric_value_vector: np.ndarray
        metric_value_range_min_idx: tuple
        metric_value_range_max_idx: tuple
        metric_value_estimation_histogram_idx: tuple
        numeric_range_estimation_result: NumericRangeEstimationResult
        for metric_value_idx in metric_value_vector_indices:
            # Obtain "N"-element-long vector of samples for each element of multi-dimensional metric.
            metric_value_vector = metric_values[metric_value_idx]
            if not datetime_detected and np.all(
                np.isclose(metric_value_vector, metric_value_vector[0])
            ):
                # Computation is unnecessary if distribution is degenerate.
                numeric_range_estimation_result = build_numeric_range_estimation_result(
                    metric_values=metric_value_vector,
                    min_value=metric_value_vector[0],
                    max_value=metric_value_vector[0],
                )
            else:
                # Compute low and high estimates for vector of samples for given element of multi-dimensional metric.
                numeric_range_estimation_result = (
                    numeric_range_estimator.get_numeric_range_estimate(
                        metric_values=metric_value_vector,
                        domain=domain,
                        variables=variables,
                        parameters=parameters,
                    )
                )

            min_value = numeric_range_estimation_result.value_range[0]
            if lower_bound is not None:
                min_value = max(np.float64(min_value), np.float64(lower_bound))

            max_value = numeric_range_estimation_result.value_range[1]
            if upper_bound is not None:
                max_value = min(np.float64(max_value), np.float64(upper_bound))

            # Obtain index of metric element (by discarding "N"-element samples dimension).
            metric_value_idx = metric_value_idx[1:]

            # Compute indices for metric value range min and max estimates.
            metric_value_range_min_idx = metric_value_idx + (
                slice(0, 1, None),
            )  # appends "[0]" element
            metric_value_range_max_idx = metric_value_idx + (
                slice(1, 2, None),
            )  # appends "[1]" element

            # Compute index for metric value estimation histogram.
            metric_value_estimation_histogram_idx = metric_value_idx

            # Store computed min and max value estimates into allocated range estimate for multi-dimensional metric.
            if datetime_detected:
                metric_value_range[metric_value_range_min_idx] = min_value
                metric_value_range[metric_value_range_max_idx] = max_value
            else:
                if round_decimals is None:
                    metric_value_range[metric_value_range_min_idx] = np.float64(
                        min_value
                    )
                    metric_value_range[metric_value_range_max_idx] = np.float64(
                        max_value
                    )
                else:
                    metric_value_range[metric_value_range_min_idx] = round(
                        np.float64(min_value), round_decimals
                    )
                    metric_value_range[metric_value_range_max_idx] = round(
                        np.float64(max_value), round_decimals
                    )

            # Store computed estimation_histogram into allocated range estimate for multi-dimensional metric.
            estimation_histogram[
                metric_value_estimation_histogram_idx
            ] = numeric_range_estimation_result.estimation_histogram

        # As a simplification, apply reduction to scalar in case of one-dimensional metric (for convenience).
        if metric_value_range.shape[0] == 1:
            metric_value_range = metric_value_range[0]
            estimation_histogram = estimation_histogram[0]

        if round_decimals == 0:
            metric_value_range = metric_value_range.astype(np.int64)

        return NumericRangeEstimationResult(
            estimation_histogram=estimation_histogram,
            value_range=metric_value_range,
        )

    @staticmethod
    def _is_metric_values_ndarray_datetime_dtype(
        metric_values: np.ndarray,
        metric_value_vector_indices: List[tuple],
    ) -> bool:
        metric_value_vector: np.ndarray
        for metric_value_idx in metric_value_vector_indices:
            metric_value_vector = metric_values[metric_value_idx]
            if not is_ndarray_datetime_dtype(
                data=metric_value_vector,
                parse_strings_as_datetimes=True,
                fuzzy=False,
            ):
                return False

        return True

    @staticmethod
    def _is_metric_values_ndarray_decimal_dtype(
        metric_values: np.ndarray,
        metric_value_vector_indices: List[tuple],
    ) -> bool:
        metric_value_vector: np.ndarray
        for metric_value_idx in metric_value_vector_indices:
            metric_value_vector = metric_values[metric_value_idx]
            if not does_ndarray_contain_decimal_dtype(data=metric_value_vector):
                return False

        return True

    def _get_truncate_values_using_heuristics(
        self,
        metric_values: np.ndarray,
        domain: Domain,
        *,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
    ) -> Dict[str, Union[Optional[int], Optional[float]]]:
        # Obtain truncate_values directive from "rule state" (i.e., variables and parameters); from instance variable otherwise.
        truncate_values: Dict[
            str, Optional[Number]
        ] = get_parameter_value_and_validate_return_type(
            domain=domain,
            parameter_reference=self.truncate_values,
            expected_return_type=dict,
            variables=variables,
            parameters=parameters,
        )

        distribution_boundary: Optional[Union[int, float]]
        if not all(
            [
                (
                    distribution_boundary is None
                    or is_numeric(value=distribution_boundary)
                    or isinstance(distribution_boundary, datetime.datetime)
                )
                for distribution_boundary in truncate_values.values()
            ]
        ):
            raise gx_exceptions.ProfilerExecutionError(
                message=f"""The directive "truncate_values" for {self.__class__.__name__} must specify the
[lower_bound, upper_bound] closed interval, where either boundary is a numeric value (or None).
"""
            )

        lower_bound: Optional[Number] = truncate_values.get("lower_bound")
        upper_bound: Optional[Number] = truncate_values.get("upper_bound")

        if metric_values.shape == 1 and np.issubdtype(metric_values.dtype, np.number):
            if lower_bound is None and np.all(np.greater(metric_values, NP_EPSILON)):
                lower_bound = 0.0

            if upper_bound is None and np.all(np.less(metric_values, (-NP_EPSILON))):
                upper_bound = 0.0

        return {
            "lower_bound": lower_bound,
            "upper_bound": upper_bound,
        }

    def _get_round_decimals_using_heuristics(
        self,
        metric_values: np.ndarray,
        domain: Domain,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
    ) -> int:
        # Obtain round_decimals directive from "rule state" (i.e., variables and parameters); from instance variable otherwise.
        round_decimals: Optional[int] = get_parameter_value_and_validate_return_type(
            domain=domain,
            parameter_reference=self.round_decimals,
            expected_return_type=None,
            variables=variables,
            parameters=parameters,
        )
        if not (
            round_decimals is None
            or (isinstance(round_decimals, int) and (round_decimals >= 0))
        ):
            raise gx_exceptions.ProfilerExecutionError(
                message=f"""The directive "round_decimals" for {self.__class__.__name__} can be 0 or a
positive integer, or must be omitted (or set to None).
"""
            )

        if np.issubdtype(metric_values.dtype, np.integer):
            round_decimals = 0

        return round_decimals



import json
import os
from collections import OrderedDict

import pytest

import great_expectations as gx
from great_expectations import DataContext
from great_expectations.core import ExpectationConfiguration, ExpectationSuite
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResultSchema,
)
from great_expectations.render import (
    RenderedDocumentContent,
    RenderedHeaderContent,
    RenderedSectionContent,
    RenderedTableContent,
)
from great_expectations.render.renderer import (
    ExpectationSuitePageRenderer,
    ValidationResultsPageRenderer,
)
from great_expectations.render.renderer_configuration import MetaNotesFormat
from great_expectations.render.view import DefaultMarkdownPageView
from great_expectations.validation_operators.types.validation_operator_result import (
    ValidationOperatorResult,
)


@pytest.fixture()
def validation_operator_result():
    fixture_filename = os.path.join(
        os.path.dirname(__file__),
        "fixtures/ValidationOperatorResult_with_multiple_validation_results.json",
    )
    with open(fixture_filename) as infile:
        validation_operator_result = json.load(infile, object_pairs_hook=OrderedDict)
        run_results = validation_operator_result["run_results"]
        for k, validation_result in run_results.items():
            validation_result[
                "validation_result"
            ] = ExpectationSuiteValidationResultSchema().load(
                validation_result["validation_result"]
            )
        return validation_operator_result


@pytest.fixture()
def expectation_suite_to_render_with_notes(empty_data_context):
    context: DataContext = empty_data_context
    expectation_suite = ExpectationSuite(
        expectation_suite_name="default",
        meta={"great_expectations_version": "0.13.0-test"},
        expectations=[
            ExpectationConfiguration(
                expectation_type="expect_column_to_exist",
                kwargs={"column": "infinities"},
            ),
            ExpectationConfiguration(
                expectation_type="expect_column_to_exist", kwargs={"column": "nulls"}
            ),
            ExpectationConfiguration(
                expectation_type="expect_column_to_exist", kwargs={"column": "naturals"}
            ),
            ExpectationConfiguration(
                expectation_type="expect_column_distinct_values_to_be_in_set",
                kwargs={"column": "irrationals", "value_set": ["*", "1st", "2nd"]},
            ),
            ExpectationConfiguration(
                expectation_type="expect_column_values_to_be_unique",
                kwargs={"column": "testings"},
                meta={
                    "notes": {
                        "content": [
                            "Example notes about this expectation. **Markdown** `Supported`.",
                            "Second example note **with** *Markdown*",
                        ],
                        "format": MetaNotesFormat.MARKDOWN,
                    }
                },
            ),
        ],
        data_context=context,
    )
    return expectation_suite


def test_render_section_page():
    section = RenderedSectionContent(
        **{
            "section_name": None,
            "content_blocks": [
                RenderedHeaderContent(
                    **{
                        "content_block_type": "header",
                        "header": "Overview",
                    }
                ),
                RenderedTableContent(
                    **{
                        "content_block_type": "table",
                        "header": "Dataset info",
                        "table": [
                            ["Number of variables", "12"],
                            ["Number of observations", "891"],
                        ],
                        "styling": {
                            "classes": ["col-6", "table-responsive"],
                            "styles": {"margin-top": "20px"},
                            "body": {"classes": ["table", "table-sm"]},
                        },
                    }
                ),
            ],
        }
    )

    rendered_doc = gx.render.view.view.DefaultMarkdownPageView().render(
        RenderedDocumentContent(sections=[section])
    )

    rendered_doc = rendered_doc.replace(" ", "").replace("\t", "").replace("\n", "")

    assert (
        rendered_doc
        == """
        #ValidationResults
        ##Overview
        ###Datasetinfo
        ||||------------|------------|Numberofvariables|12Numberofobservations|891
        -----------------------------------------------------------
        Poweredby[GreatExpectations](https://greatexpectations.io/)
        """.replace(
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )


def test_snapshot_render_section_page_with_fixture_data(validation_operator_result):
    """
    Make sure the appropriate markdown rendering is done for the applied fixture.
    Args:
        validation_operator_result: test fixture

    Returns: None

    """
    validation_operator_result = ValidationOperatorResult(**validation_operator_result)

    validation_results_page_renderer = ValidationResultsPageRenderer(
        run_info_at_end=True
    )

    rendered_document_content_list = (
        validation_results_page_renderer.render_validation_operator_result(
            validation_operator_result=validation_operator_result
        )
    )

    md_str_list = DefaultMarkdownPageView().render(rendered_document_content_list)

    md_str = " ".join(md_str_list)

    md_str = md_str.replace(" ", "").replace("\t", "").replace("\n", "")

    assert (
        md_str
        == """
# Validation Results
## Overview
### **Expectation Suite:** **basic.warning**
**Data asset:** **None**
**Status:**  **Failed**
### Statistics
 |  |  |
 | ------------  | ------------ |
Evaluated Expectations  | 11
Successful Expectations  | 9
Unsuccessful Expectations  | 2
Success Percent  | ≈81.82%
## Table-Level Expectations
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
❌  | Must have greater than or equal to **27000** and less than or equal to **33000** rows.  | 30
✅  | Must have exactly **3** columns.  | 3
✅  | Must have these columns in this order: **Team**, ** "Payroll (millions)"**, ** "Wins"**  | ['Team', ' "Payroll (millions)"', ' "Wins"']
##  "Payroll (millions)"
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
✅  | values must never be null.  | 100% not null
✅  | minimum value must be greater than or equal to **54.24** and less than or equal to **56.24**.  | 55.24
✅  | maximum value must be greater than or equal to **196.96** and less than or equal to **198.96**.  | 197.96
✅  | mean must be greater than or equal to **97.01899999999998** and less than or equal to **99.01899999999998**.  | ≈98.019
❌  | median must be greater than or equal to **84000.75** and less than or equal to **86000.75**.  | 85.75
✅  | quantiles must be within the following value ranges.
 | Quantile | Min Value | Max Value |
 | ------------  | ------------  | ------------ |
0.05  | 54.37  | 56.37
Q1  | 74.48  | 76.48
Median  | 82.31  | 84.31
Q3  | 116.62  | 118.62
0.95  | 173.54  | 175.54
  |
 | Quantile | Value |
 | ------------  | ------------ |
0.05  | 55.37
Q1  | 75.48
Median  | 83.31
Q3  | 117.62
0.95  | 174.54
## Team
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
✅  | values must never be null.  | 100% not null
✅  | values must always be greater than or equal to **1** characters long.  | 0% unexpected
### Info
 |  |  |
 | ------------  | ------------ |
Great Expectations Version  | 0.11.8+4.g4ab34df3.dirty
Run Name  | getest run
Run Time  | 2020-07-27T17:19:32Z
### Batch Markers
 |  |  |
 | ------------  | ------------ |
**ge_load_time**  | **20200727T171932.954810Z**
**pandas_data_fingerprint**  | **8c46fdaf0bd356fd58b7bcd9b2e6012d**
### Batch Kwargs
 |  |  |
 | ------------  | ------------ |
**PandasInMemoryDF**  | **True**
**datasource**  | **getest**
**ge_batch_id**  | **56615f40-d02d-11ea-b6ea-acde48001122**
-----------------------------------------------------------
Powered by [Great Expectations](https://greatexpectations.io/)
# Validation Results
## Overview
### **Expectation Suite:** **basic.warning**
**Data asset:** **None**
**Status:**  **Failed**
### Statistics
 |  |  |
 | ------------  | ------------ |
Evaluated Expectations  | 11
Successful Expectations  | 9
Unsuccessful Expectations  | 2
Success Percent  | ≈81.82%
## Table-Level Expectations
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
❌  | Must have greater than or equal to **27000** and less than or equal to **33000** rows.  | 30
✅  | Must have exactly **3** columns.  | 3
✅  | Must have these columns in this order: **Team**, ** "Payroll (millions)"**, ** "Wins"**  | ['Team', ' "Payroll (millions)"', ' "Wins"']
##  "Payroll (millions)"
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
✅  | values must never be null.  | 100% not null
✅  | minimum value must be greater than or equal to **54.24** and less than or equal to **56.24**.  | 55.24
✅  | maximum value must be greater than or equal to **196.96** and less than or equal to **198.96**.  | 197.96
✅  | mean must be greater than or equal to **97.01899999999998** and less than or equal to **99.01899999999998**.  | ≈98.019
❌  | median must be greater than or equal to **84000.75** and less than or equal to **86000.75**.  | 85.75
✅  | quantiles must be within the following value ranges.
 | Quantile | Min Value | Max Value |
 | ------------  | ------------  | ------------ |
0.05  | 54.37  | 56.37
Q1  | 74.48  | 76.48
Median  | 82.31  | 84.31
Q3  | 116.62  | 118.62
0.95  | 173.54  | 175.54
  |
 | Quantile | Value |
 | ------------  | ------------ |
0.05  | 55.37
Q1  | 75.48
Median  | 83.31
Q3  | 117.62
0.95  | 174.54
## Team
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
✅  | values must never be null.  | 100% not null
✅  | values must always be greater than or equal to **1** characters long.  | 0% unexpected
### Info
 |  |  |
 | ------------  | ------------ |
Great Expectations Version  | 0.11.8+4.g4ab34df3.dirty
Run Name  | getest run
Run Time  | 2020-07-27T17:19:32Z
### Batch Markers
 |  |  |
 | ------------  | ------------ |
**ge_load_time**  | **20200727T171932.954810Z**
**pandas_data_fingerprint**  | **8c46fdaf0bd356fd58b7bcd9b2e6012d**
### Batch Kwargs
 |  |  |
 | ------------  | ------------ |
**PandasInMemoryDF**  | **True**
**datasource**  | **getest**
**ge_batch_id**  | **56615f40-d02d-11ea-b6ea-acde48001122**
-----------------------------------------------------------
Powered by [Great Expectations](https://greatexpectations.io/)
""".replace(
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )


def test_render_section_page_with_fixture_data_multiple_validations(
    validation_operator_result,
):
    """
    Make sure the appropriate markdown rendering is done for the applied fixture.
    :param validation_operator_result: test fixture
    :return: None
    """

    validation_operator_result = ValidationOperatorResult(**validation_operator_result)

    validation_results_page_renderer = ValidationResultsPageRenderer(
        run_info_at_end=True
    )

    rendered_document_content_list = (
        validation_results_page_renderer.render_validation_operator_result(
            validation_operator_result=validation_operator_result
        )
    )

    md_str_list = DefaultMarkdownPageView().render(rendered_document_content_list)

    md_str = " ".join(md_str_list)

    md_str = md_str.replace(" ", "").replace("\t", "").replace("\n", "")

    # print(md_str)

    assert (
        md_str
        == """
# Validation Results
## Overview
### **Expectation Suite:** **basic.warning**
**Dataasset:** **None**
**Status:**  **Failed**
### Statistics
 |  |  |
 | ------------  | ------------ |
Evaluated Expectations  | 11
Successful Expectations  | 9
Unsuccessful Expectations  | 2
Success Percent  | ≈81.82%
## Table-Level Expectations
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
❌  | Must have greater than or equal to **27000** and less than or equal to **33000** rows.  | 30
✅  | Must have exactly **3** columns.  | 3
✅  | Must have these columns in this order: **Team**, ** "Payroll (millions)"**, ** "Wins"**  | ['Team', ' "Payroll (millions)"', ' "Wins"']
##  "Payroll (millions)"
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
✅  | values must never be null.  | 100% not null
✅  | minimum value must be greater than or equal to **54.24** and less than or equal to **56.24**.  | 55.24
✅  | maximum value must be greater than or equal to **196.96** and less than or equal to **198.96**.  | 197.96
✅  | mean must be greater than or equal to **97.01899999999998** and less than or equal to **99.01899999999998**.  | ≈98.019
❌  | median must be greater than or equal to **84000.75** and less than or equal to **86000.75**.  | 85.75
✅  | quantiles must be within the following value ranges.
 | Quantile | Min Value | Max Value |
 | ------------  | ------------  | ------------ |
0.05  | 54.37  | 56.37
Q1  | 74.48  | 76.48
Median  | 82.31  | 84.31
Q3  | 116.62  | 118.62
0.95  | 173.54  | 175.54
  |
 | Quantile | Value |
 | ------------  | ------------ |
0.05  | 55.37
Q1  | 75.48
Median  | 83.31
Q3  | 117.62
0.95  | 174.54
## Team
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
✅  | values must never be null.  | 100% not null
✅  | values must always be greater than or equal to **1** characters long.  | 0% unexpected
### Info
 |  |  |
 | ------------  | ------------ |
Great Expectations Version  | 0.11.8+4.g4ab34df3.dirty
Run Name  | getest run
Run Time  | 2020-07-27T17:19:32Z
### Batch Markers
 |  |  |
 | ------------  | ------------ |
**ge_load_time**  | **20200727T171932.954810Z**
**pandas_data_fingerprint**  | **8c46fdaf0bd356fd58b7bcd9b2e6012d**
### Batch Kwargs
 |  |  |
 | ------------  | ------------ |
**PandasInMemoryDF**  | **True**
**datasource**  | **getest**
**ge_batch_id**  | **56615f40-d02d-11ea-b6ea-acde48001122**
-----------------------------------------------------------
Powered by [Great Expectations](https://greatexpectations.io/)
# Validation Results
## Overview
### **Expectation Suite:** **basic.warning**
**Dataasset:** **None**
**Status:**  **Failed**
### Statistics
 |  |  |
 | ------------  | ------------ |
Evaluated Expectations  | 11
Successful Expectations  | 9
Unsuccessful Expectations  | 2
Success Percent  | ≈81.82%
## Table-Level Expectations
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
❌  | Must have greater than or equal to **27000** and less than or equal to **33000** rows.  | 30
✅  | Must have exactly **3** columns.  | 3
✅  | Must have these columns in this order: **Team**, ** "Payroll (millions)"**, ** "Wins"**  | ['Team', ' "Payroll (millions)"', ' "Wins"']
##  "Payroll (millions)"
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
✅  | values must never be null.  | 100% not null
✅  | minimum value must be greater than or equal to **54.24** and less than or equal to **56.24**.  | 55.24
✅  | maximum value must be greater than or equal to **196.96** and less than or equal to **198.96**.  | 197.96
✅  | mean must be greater than or equal to **97.01899999999998** and less than or equal to **99.01899999999998**.  | ≈98.019
❌  | median must be greater than or equal to **84000.75** and less than or equal to **86000.75**.  | 85.75
✅  | quantiles must be within the following value ranges.
 | Quantile | Min Value | Max Value |
 | ------------  | ------------  | ------------ |
0.05  | 54.37  | 56.37
Q1  | 74.48  | 76.48
Median  | 82.31  | 84.31
Q3  | 116.62  | 118.62
0.95  | 173.54  | 175.54
  |
 | Quantile | Value |
 | ------------  | ------------ |
0.05  | 55.37
Q1  | 75.48
Median  | 83.31
Q3  | 117.62
0.95  | 174.54
## Team
 | Status | Expectation | Observed Value |
 | ------------  | ------------  | ------------ |
✅  | values must never be null.  | 100% not null
✅  | values must always be greater than or equal to **1** characters long.  | 0% unexpected
### Info
 |  |  |
 | ------------  | ------------ |
Great Expectations Version  | 0.11.8+4.g4ab34df3.dirty
Run Name  | getest run
Run Time  | 2020-07-27T17:19:32Z
### Batch Markers
 |  |  |
 | ------------  | ------------ |
**ge_load_time**  | **20200727T171932.954810Z**
**pandas_data_fingerprint**  | **8c46fdaf0bd356fd58b7bcd9b2e6012d**
### Batch Kwargs
 |  |  |
 | ------------  | ------------ |
**PandasInMemoryDF**  | **True**
**datasource**  | **getest**
**ge_batch_id**  | **56615f40-d02d-11ea-b6ea-acde48001122**
-----------------------------------------------------------
Powered by [Great Expectations](https://greatexpectations.io/)
""".replace(
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )


def test_render_expectation_suite_for_Markdown(expectation_suite_to_render_with_notes):
    expectation_suite_page_renderer = ExpectationSuitePageRenderer()
    rendered_document_content_list = expectation_suite_page_renderer.render(
        expectation_suite_to_render_with_notes
    )
    md_str = DefaultMarkdownPageView().render(rendered_document_content_list)
    md_str = " ".join(md_str)
    md_str = md_str.replace(" ", "").replace("\t", "").replace("\n", "")
    assert (
        md_str
        == r"""
   # Validation Results
## Overview
### Info
 |  |  |
 | ------------  | ------------ |
Expectation Suite Name  | default
Great Expectations Version  | 0.13.0-test
### Notes
    This Expectation suite currently contains 5 total Expectations across 5 columns.
## infinities
  * is a required field.
  * ***
## irrationals
  * distinct values must belong to this set: \* **1st** **2nd**.
  * ***
## naturals
  * is a required field.
  * ***
## nulls
  * is a required field.
  * ***
## testings
  * values must be unique.
    #### Notes:
      Example notes about this expectation. **Markdown** `Supported`.

      Second example note **with** *Markdown*
  * ***
-----------------------------------------------------------
Powered by [Great Expectations](https://greatexpectations.io/)
    """.replace(
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from great_expectations.core.domain import Domain
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.rule_based_profiler.domain_builder import DomainBuilder
from great_expectations.rule_based_profiler.helpers.util import (
    get_parameter_value_and_validate_return_type,
)
from great_expectations.rule_based_profiler.parameter_container import (
    VARIABLES_KEY,
    ParameterContainer,
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )


class TableDomainBuilder(DomainBuilder):
    def __init__(
        self,
        data_context: Optional[AbstractDataContext] = None,
    ) -> None:
        """
        Args:
            data_context: AbstractDataContext associated with this DomainBuilder
        """
        super().__init__(data_context=data_context)

    @property
    def domain_type(self) -> MetricDomainTypes:
        return MetricDomainTypes.TABLE

    """
    The interface method of TableDomainBuilder emits a single Domain object, corresponding to the implied Batch (table).

    Note that for appropriate use-cases, it should be readily possible to build a multi-batch implementation, where a
    separate Domain object is emitted for each individual Batch (using its respective batch_id).  (This is future work.)
    """

    def _get_domains(
        self,
        rule_name: str,
        variables: Optional[ParameterContainer] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> List[Domain]:
        other_table_name: Optional[str]
        try:
            # Obtain table from "rule state" (i.e., variables and parameters); from instance variable otherwise.
            other_table_name = get_parameter_value_and_validate_return_type(
                domain=None,
                parameter_reference=f"{VARIABLES_KEY}table",
                expected_return_type=None,
                variables=variables,
                parameters=None,
            )
        except KeyError:
            other_table_name = None

        domains: List[Domain]
        if other_table_name:
            domains = [
                Domain(
                    domain_type=self.domain_type,
                    domain_kwargs={
                        "table": other_table_name,
                    },
                    rule_name=rule_name,
                ),
            ]
        else:
            domains = [
                Domain(
                    domain_type=self.domain_type,
                    rule_name=rule_name,
                ),
            ]

        return domains



import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Type

from great_expectations.core.usage_statistics.anonymizers.base import BaseAnonymizer

if TYPE_CHECKING:
    from great_expectations.core import ExpectationSuite
    from great_expectations.data_context.store import Store
    from great_expectations.validation_operators import ValidationOperator

logger = logging.getLogger(__name__)


class Anonymizer(BaseAnonymizer):
    def __init__(self, salt: Optional[str] = None) -> None:
        super().__init__(salt=salt)

        from great_expectations.core.usage_statistics.anonymizers.action_anonymizer import (
            ActionAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.batch_anonymizer import (
            BatchAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.batch_request_anonymizer import (
            BatchRequestAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.checkpoint_anonymizer import (
            CheckpointAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.data_connector_anonymizer import (
            DataConnectorAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.data_docs_anonymizer import (
            DataDocsAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.datasource_anonymizer import (
            DatasourceAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.expectation_anonymizer import (
            ExpectationSuiteAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.profiler_anonymizer import (
            ProfilerAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.store_anonymizer import (
            StoreAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.store_backend_anonymizer import (
            StoreBackendAnonymizer,
        )
        from great_expectations.core.usage_statistics.anonymizers.validation_operator_anonymizer import (
            ValidationOperatorAnonymizer,
        )

        self._strategies: List[Type[BaseAnonymizer]] = [
            CheckpointAnonymizer,
            ProfilerAnonymizer,
            DatasourceAnonymizer,
            DataConnectorAnonymizer,
            ActionAnonymizer,
            DataDocsAnonymizer,
            ExpectationSuiteAnonymizer,
            ValidationOperatorAnonymizer,
            BatchRequestAnonymizer,
            BatchAnonymizer,
            StoreAnonymizer,
            StoreBackendAnonymizer,
        ]

        # noinspection PyArgumentList
        self._anonymizers: Dict[Type[BaseAnonymizer], BaseAnonymizer] = {
            strategy: strategy(salt=self._salt, aggregate_anonymizer=self)
            for strategy in self._strategies
        }

    def anonymize(self, obj: Optional[object] = None, **kwargs: Optional[dict]) -> Any:
        anonymizer: Optional[BaseAnonymizer] = self._get_anonymizer(obj=obj, **kwargs)

        if anonymizer is not None:
            return anonymizer.anonymize(obj=obj, **kwargs)
        elif isinstance(obj, str):
            return self._anonymize_string(string_=obj)
        elif not obj:
            return obj

        raise TypeError(
            f"The type {type(obj)} cannot be handled by the Anonymizer; no suitable strategy found."
        )

    def can_handle(self, obj: object, **kwargs: Optional[dict]) -> bool:
        return self._get_anonymizer(obj=obj, **kwargs) is not None

    def _get_anonymizer(self, obj: object, **kwargs) -> Optional[BaseAnonymizer]:
        for anonymizer in self._anonymizers.values():
            if anonymizer.can_handle(obj=obj, **kwargs):
                return anonymizer

        return None

    def anonymize_init_payload(self, init_payload: Dict[str, Any]) -> Dict[str, Any]:
        anonymized_init_payload: Dict[str, Any] = {}
        anonymizer_funcs: Dict[str, Any] = {
            "datasources": self._anonymize_datasources_init_payload,
            "stores": self._anonymize_stores_init_payload,
            "validation_operators": self._anonymize_validation_operator_init_payload,
            "data_docs_sites": self._anonymize_data_docs_sites_init_payload,
            "expectation_suites": self._anonymize_expectation_suite_init_payload,
            "dependencies": None,  # dependencies do not need anonymization
        }

        for key, val in init_payload.items():
            anonymizer_func: Optional[Callable] = anonymizer_funcs.get(key)
            if anonymizer_func:
                anonymized_key: str = f"anonymized_{key}"
                anonymized_init_payload[anonymized_key] = anonymizer_func(val)
            else:
                anonymized_init_payload[key] = val
        return anonymized_init_payload

    def _anonymize_datasources_init_payload(self, payload: dict) -> List[dict]:
        from great_expectations.core.usage_statistics.anonymizers.datasource_anonymizer import (
            DatasourceAnonymizer,
        )

        anonymizer = self._anonymizers[DatasourceAnonymizer]

        anonymized_values: List[dict] = []
        for name, config in payload.items():
            anonymize_value: dict = anonymizer._anonymize_datasource_info(
                name=name, config=config
            )
            anonymized_values.append(anonymize_value)

        return anonymized_values

    def _anonymize_stores_init_payload(self, payload: Dict[str, "Store"]) -> List[dict]:
        from great_expectations.core.usage_statistics.anonymizers.store_anonymizer import (
            StoreAnonymizer,
        )

        anonymizer = self._anonymizers[StoreAnonymizer]

        anonymized_values: List[dict] = []
        for store_name, store_obj in payload.items():
            anonymize_value: dict = anonymizer.anonymize(
                store_name=store_name,
                store_obj=store_obj,
            )
            anonymized_values.append(anonymize_value)

        return anonymized_values

    def _anonymize_validation_operator_init_payload(
        self, payload: Optional[Dict[str, "ValidationOperator"]] = None
    ) -> List[dict]:

        from great_expectations.core.usage_statistics.anonymizers.validation_operator_anonymizer import (
            ValidationOperatorAnonymizer,
        )

        if payload is None:
            return []

        anonymizer = self._anonymizers[ValidationOperatorAnonymizer]

        anonymized_values: List[dict] = []
        for validation_operator_name, validation_operator_obj in payload.items():
            anonymize_value: dict = anonymizer.anonymize(
                validation_operator_name=validation_operator_name,
                validation_operator_obj=validation_operator_obj,
            )
            anonymized_values.append(anonymize_value)

        return anonymized_values

    def _anonymize_data_docs_sites_init_payload(
        self, payload: Dict[str, dict]
    ) -> List[dict]:
        from great_expectations.core.usage_statistics.anonymizers.data_docs_anonymizer import (
            DataDocsAnonymizer,
        )

        anonymizer = self._anonymizers[DataDocsAnonymizer]

        anonymized_values: List[dict] = []
        for site_name, site_config in payload.items():
            anonymize_value: dict = anonymizer.anonymize(
                site_name=site_name, site_config=site_config
            )
            anonymized_values.append(anonymize_value)

        return anonymized_values

    def _anonymize_expectation_suite_init_payload(
        self, payload: List["ExpectationSuite"]
    ) -> List[dict]:
        from great_expectations.core.usage_statistics.anonymizers.expectation_anonymizer import (
            ExpectationSuiteAnonymizer,
        )

        anonymizer = self._anonymizers[ExpectationSuiteAnonymizer]

        anonymized_values: List[dict] = []
        for suite in payload:
            anonymize_value: dict = anonymizer.anonymize(obj=suite)
            anonymized_values.append(anonymize_value)

        return anonymized_values



"""
Test GXCloudStoreBackend behavior and adherence to StoreBackend contract.

Since GXCloudStoreBackend relies on GX Cloud, we mock requests and assert the right calls
are made from the Store API (set, get, list, and remove_key).

Note that although ge_cloud_access_token is provided (and is a valid UUID), no external
requests are actually made as part of this test. The actual value of the token does not
matter here but we leverage an existing fixture to mimic the contents of requests made
in production. The same logic applies to all UUIDs in this test.
"""

from collections import OrderedDict
from typing import Callable, Optional, Set, Union
from unittest import mock

import pytest

from great_expectations.data_context.cloud_constants import (
    CLOUD_DEFAULT_BASE_URL,
    GXCloudRESTResource,
)
from great_expectations.data_context.store.gx_cloud_store_backend import (
    GXCloudStoreBackend,
    construct_json_payload,
    construct_url,
)
from great_expectations.data_context.types.base import CheckpointConfig


@pytest.fixture
def construct_ge_cloud_store_backend(
    ge_cloud_access_token: str,
) -> Callable[[GXCloudRESTResource], GXCloudStoreBackend]:
    def _closure(resource_type: GXCloudRESTResource) -> GXCloudStoreBackend:
        ge_cloud_base_url = CLOUD_DEFAULT_BASE_URL
        ge_cloud_credentials = {
            "access_token": ge_cloud_access_token,
            "organization_id": "51379b8b-86d3-4fe7-84e9-e1a52f4a414c",
        }

        store_backend = GXCloudStoreBackend(
            ge_cloud_base_url=ge_cloud_base_url,
            ge_cloud_credentials=ge_cloud_credentials,
            ge_cloud_resource_type=resource_type,
        )
        return store_backend

    return _closure


@pytest.mark.parametrize(
    "base_url,organization_id,resource_name,id,expected",
    [
        pytest.param(
            "https://app.test.greatexpectations.io",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_resource",
            None,
            "https://app.test.greatexpectations.io/organizations/de5b9ca6-caf7-43c8-a820-5540ec6df9b2/my-resource",
            id="no id",
        ),
        pytest.param(
            "https://app.test.greatexpectations.io",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_resource",
            "8746e25c-de4d-450d-967b-df0d5546590d",
            "https://app.test.greatexpectations.io/organizations/de5b9ca6-caf7-43c8-a820-5540ec6df9b2/my-resource/8746e25c-de4d-450d-967b-df0d5546590d",
            id="with id",
        ),
    ],
)
@pytest.mark.unit
@pytest.mark.cloud
def test_construct_url(
    base_url: str,
    organization_id: str,
    resource_name: str,
    id: Optional[str],
    expected: str,
) -> None:
    assert (
        construct_url(
            base_url=base_url,
            organization_id=organization_id,
            resource_name=resource_name,
            id=id,
        )
        == expected
    )


@pytest.mark.parametrize(
    "resource_type,organization_id,attributes_key,attributes_value,kwargs,expected",
    [
        pytest.param(
            "my_resource",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_attribute",
            "my_value",
            {},
            {
                "data": {
                    "type": "my_resource",
                    "attributes": {
                        "organization_id": "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
                        "my_attribute": "my_value",
                    },
                },
            },
            id="no kwargs",
        ),
        pytest.param(
            "my_resource",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_attribute",
            {"key1": {"nested_key1": 1}, "key2": {"nested_key2": 2}},
            {},
            {
                "data": {
                    "type": "my_resource",
                    "attributes": {
                        "organization_id": "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
                        "my_attribute": {
                            "key1": {"nested_key1": 1},
                            "key2": {"nested_key2": 2},
                        },
                    },
                },
            },
            id="with nested value",
        ),
        pytest.param(
            "my_resource",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_attribute",
            "my_value",
            {
                "kwarg1": 1,
                "kwarg2": 2,
            },
            {
                "data": {
                    "type": "my_resource",
                    "attributes": {
                        "organization_id": "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
                        "my_attribute": "my_value",
                        "kwarg1": 1,
                        "kwarg2": 2,
                    },
                },
            },
            id="with basic kwargs",
        ),
        pytest.param(
            "my_resource",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_attribute",
            "my_value",
            {
                "kwarg1": {
                    "nested1": 1,
                },
                "kwarg2": {
                    "nested2": 2,
                },
            },
            {
                "data": {
                    "type": "my_resource",
                    "attributes": {
                        "organization_id": "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
                        "my_attribute": "my_value",
                        "kwarg1": {
                            "nested1": 1,
                        },
                        "kwarg2": {
                            "nested2": 2,
                        },
                    },
                },
            },
            id="with nested kwargs",
        ),
    ],
)
@pytest.mark.unit
@pytest.mark.cloud
def test_construct_json_payload(
    resource_type: str,
    organization_id: str,
    attributes_key: str,
    attributes_value: str,
    kwargs: dict,
    expected: dict,
) -> None:
    assert (
        construct_json_payload(
            resource_type=resource_type,
            organization_id=organization_id,
            attributes_key=attributes_key,
            attributes_value=attributes_value,
            **kwargs,
        )
        == expected
    )


@pytest.mark.cloud
@pytest.mark.unit
def test_set(
    construct_ge_cloud_store_backend: Callable[
        [GXCloudRESTResource], GXCloudStoreBackend
    ],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    my_simple_checkpoint_config: CheckpointConfig = CheckpointConfig(
        name="my_minimal_simple_checkpoint",
        class_name="SimpleCheckpoint",
        config_version=1,
    )
    my_simple_checkpoint_config_serialized = (
        my_simple_checkpoint_config.get_schema_class()().dump(
            my_simple_checkpoint_config
        )
    )

    with mock.patch("requests.Session.post", autospec=True) as mock_post:
        store_backend.set(("checkpoint", ""), my_simple_checkpoint_config_serialized)
        mock_post.assert_called_with(
            mock.ANY,  # requests.Session object
            f"{CLOUD_DEFAULT_BASE_URL}organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/checkpoints",
            json={
                "data": {
                    "type": "checkpoint",
                    "attributes": {
                        "organization_id": "51379b8b-86d3-4fe7-84e9-e1a52f4a414c",
                        "checkpoint_config": OrderedDict(
                            [
                                ("name", "my_minimal_simple_checkpoint"),
                                ("config_version", 1.0),
                                ("template_name", None),
                                ("module_name", "great_expectations.checkpoint"),
                                ("class_name", "SimpleCheckpoint"),
                                ("run_name_template", None),
                                ("expectation_suite_name", None),
                                ("batch_request", {}),
                                ("action_list", []),
                                ("evaluation_parameters", {}),
                                ("runtime_configuration", {}),
                                ("validations", []),
                                ("profilers", []),
                                ("ge_cloud_id", None),
                                ("expectation_suite_ge_cloud_id", None),
                            ]
                        ),
                    },
                }
            },
        )


@pytest.mark.cloud
@pytest.mark.unit
def test_list_keys(
    construct_ge_cloud_store_backend: Callable[
        [GXCloudRESTResource], GXCloudStoreBackend
    ],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    with mock.patch("requests.Session.get", autospec=True) as mock_get:
        store_backend.list_keys()
        mock_get.assert_called_with(
            mock.ANY,  # requests.Session object
            f"{CLOUD_DEFAULT_BASE_URL}organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/checkpoints",
        )


@pytest.mark.cloud
@pytest.mark.unit
def test_remove_key(
    construct_ge_cloud_store_backend: Callable[
        [GXCloudRESTResource], GXCloudStoreBackend
    ],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    with mock.patch("requests.Session.delete", autospec=True) as mock_delete:
        mock_response = mock_delete.return_value
        mock_response.status_code = 200

        store_backend.remove_key(
            (
                "checkpoint",
                "0ccac18e-7631-4bdd-8a42-3c35cce574c6",
            )
        )
        mock_delete.assert_called_with(
            mock.ANY,  # requests.Session object
            f"{CLOUD_DEFAULT_BASE_URL}organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/checkpoints/0ccac18e-7631"
            "-4bdd"
            "-8a42-3c35cce574c6",
            json={
                "data": {
                    "type": "checkpoint",
                    "id": "0ccac18e-7631-4bdd-8a42-3c35cce574c6",
                    "attributes": {"deleted": True},
                }
            },
        )


@pytest.mark.cloud
@pytest.mark.unit
def test_appropriate_casting_of_str_resource_type_to_GXCloudRESTResource(
    construct_ge_cloud_store_backend: Callable[
        [GXCloudRESTResource], GXCloudStoreBackend
    ],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    assert store_backend.ge_cloud_resource_type is GXCloudRESTResource.CHECKPOINT


@pytest.mark.parametrize(
    "resource_type,expected_set_kwargs",
    [
        pytest.param(
            GXCloudRESTResource.VALIDATION_RESULT,
            {"checkpoint_id", "expectation_suite_id"},
            id="resource_with_allowed_set_kwargs",
        ),
        pytest.param(
            GXCloudRESTResource.CHECKPOINT,
            set(),
            id="resource_without_allowed_set_kwargs",
        ),
    ],
)
@pytest.mark.cloud
@pytest.mark.unit
def test_allowed_set_kwargs(
    resource_type: GXCloudRESTResource,
    expected_set_kwargs: Set[str],
    construct_ge_cloud_store_backend: Callable[
        [GXCloudRESTResource], GXCloudStoreBackend
    ],
) -> None:
    store_backend = construct_ge_cloud_store_backend(resource_type)
    actual = store_backend.allowed_set_kwargs

    assert actual == expected_set_kwargs


@pytest.mark.parametrize(
    "kwargs,expected",
    [
        pytest.param(
            {},
            True,
            id="no kwargs",
        ),
        pytest.param(
            {"checkpoint_id": "123", "expectation_suite_id": "456"},
            True,
            id="equivalent kwargs",
        ),
        pytest.param(
            {
                "checkpoint_id": "123",
                "expectation_suite_id": "456",
                "my_other_id": "789",
            },
            None,
            id="additional kwargs",
            marks=pytest.mark.xfail(strict=True, raises=ValueError),
        ),
    ],
)
@pytest.mark.cloud
@pytest.mark.unit
def test_validate_set_kwargs(
    kwargs: dict,
    expected: Union[bool, None],
    construct_ge_cloud_store_backend: Callable[
        [GXCloudRESTResource], GXCloudStoreBackend
    ],
) -> None:
    store_backend = construct_ge_cloud_store_backend(
        GXCloudRESTResource.VALIDATION_RESULT
    )
    store_backend.validate_set_kwargs(kwargs) == expected


@pytest.mark.cloud
@pytest.mark.unit
def test_config_property_and_defaults(
    construct_ge_cloud_store_backend: Callable[
        [GXCloudRESTResource], GXCloudStoreBackend
    ],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    assert store_backend.config == {
        "class_name": GXCloudStoreBackend.__name__,
        "fixed_length_key": True,
        "ge_cloud_base_url": CLOUD_DEFAULT_BASE_URL,
        "ge_cloud_resource_type": GXCloudRESTResource.CHECKPOINT,
        "manually_initialize_store_backend_id": "",
        "module_name": GXCloudStoreBackend.__module__,
        "suppress_store_backend_id": True,
    }



"""Example Script: How to create an Expectation Suite with the Onboarding Data Assistant

This example script is intended for use in documentation on how to use an Onboarding Data Assistant to create
an Expectation Suite.

Assert statements are included to ensure that if the behaviour shown in this script breaks it will not pass
tests and will be updated.  These statements can be ignored by users.

Comments with the tags `<snippet>` and `</snippet>` are used to ensure that if this script is updated
the snippets that are specified for use in documentation are maintained.  These comments can be ignored by users.

--documentation--
    https://docs.greatexpectations.io/docs/guides/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant
"""
import great_expectations as gx
from great_expectations.checkpoint import SimpleCheckpoint
from great_expectations.core.batch import BatchRequest
from great_expectations.core.yaml_handler import YAMLHandler

yaml = YAMLHandler()

context: gx.DataContext = gx.get_context()

# Configure your datasource (if you aren't using one that already exists)

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py datasource_config">
datasource_config = {
    "name": "taxi_multi_batch_datasource",
    "class_name": "Datasource",
    "module_name": "great_expectations.datasource",
    "execution_engine": {
        "module_name": "great_expectations.execution_engine",
        "class_name": "PandasExecutionEngine",
    },
    "data_connectors": {
        "inferred_data_connector_all_years": {
            "class_name": "InferredAssetFilesystemDataConnector",
            "base_directory": "<PATH_TO_YOUR_DATA_HERE>",
            "default_regex": {
                "group_names": ["data_asset_name", "year", "month"],
                "pattern": "(yellow_tripdata_sample)_(\\d.*)-(\\d.*)\\.csv",
            },
        },
    },
}
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
datasource_config["data_connectors"]["inferred_data_connector_all_years"][
    "base_directory"
] = "../data/"

context.test_yaml_config(yaml.dump(datasource_config))

# add_datasource only if it doesn't already exist in our configuration

try:
    context.get_datasource(datasource_config["name"])
except ValueError:
    context.add_datasource(**datasource_config)

# Prepare an Expectation Suite

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py expectation_suite">
expectation_suite_name = "my_onboarding_assistant_suite"

expectation_suite = context.add_or_update_expectation_suite(
    expectation_suite_name=expectation_suite_name
)
# </snippet>

# Prepare a Batch Request

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py batch_request">
multi_batch_all_years_batch_request: BatchRequest = BatchRequest(
    datasource_name="taxi_multi_batch_datasource",
    data_connector_name="inferred_data_connector_all_years",
    data_asset_name="yellow_tripdata_sample",
)
# </snippet>

# Run the Onboarding Assistant

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py exclude_column_names">
exclude_column_names = [
    "VendorID",
    "pickup_datetime",
    "dropoff_datetime",
    "RatecodeID",
    "PULocationID",
    "DOLocationID",
    "payment_type",
    "fare_amount",
    "extra",
    "mta_tax",
    "tip_amount",
    "tolls_amount",
    "improvement_surcharge",
    "congestion_surcharge",
]
# </snippet>

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py data_assistant_result">
data_assistant_result = context.assistants.onboarding.run(
    batch_request=multi_batch_all_years_batch_request,
    exclude_column_names=exclude_column_names,
)
# </snippet>

# Save your Expectation Suite

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py get_expectation_suite">
expectation_suite = data_assistant_result.get_expectation_suite(
    expectation_suite_name=expectation_suite_name
)
# </snippet>

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py save_expectation_suite">
context.add_expectation_suite(expectation_suite=expectation_suite)
# </snippet>

# Use a SimpleCheckpoint to verify that your new Expectation Suite works.

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py checkpoint_config">
checkpoint_config = {
    "class_name": "SimpleCheckpoint",
    "validations": [
        {
            "batch_request": multi_batch_all_years_batch_request,
            "expectation_suite_name": expectation_suite_name,
        }
    ],
}
# </snippet>

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py checkpoint">
checkpoint = SimpleCheckpoint(
    f"yellow_tripdata_sample_{expectation_suite_name}",
    context,
    **checkpoint_config,
)
checkpoint_result = checkpoint.run()

assert checkpoint_result["success"] is True
# </snippet>

# If you are using code from this script as part of a Jupyter Notebook, uncommenting and running the
# following lines will open your Data Docs for the `checkpoint`'s results:

# context.build_data_docs()
# validation_result_identifier = checkpoint_result.list_validation_result_identifiers()[0]
# context.open_data_docs(resource_identifier=validation_result_identifier)


# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py plot_metrics">
data_assistant_result.plot_metrics()
# </snippet>

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py metrics_by_domain">
data_assistant_result.metrics_by_domain
# </snippet>

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py plot_expectations_and_metrics">
data_assistant_result.plot_expectations_and_metrics()
# </snippet>

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py show_expectations_by_domain_type">
data_assistant_result.show_expectations_by_domain_type()
# </snippet>

# <snippet name="tests/integration/docusaurus/expectations/data_assistants/how_to_create_an_expectation_suite_with_the_onboarding_data_assistant.py show_expectations_by_expectation_type">
data_assistant_result.show_expectations_by_expectation_type()
# </snippet>



from great_expectations.data_context.types.base import (
    ConcurrencyConfig,
    DataContextConfig,
    InMemoryStoreBackendDefaults,
)
from great_expectations.util import get_context


def test_concurrency_disabled_by_default():
    data_context_config = DataContextConfig()
    assert data_context_config.concurrency is None


def test_concurrency_enabled_with_dict():
    data_context_config = DataContextConfig(concurrency={"enabled": True})
    assert data_context_config.concurrency.enabled


def test_concurrency_enabled_with_config():
    data_context_config = DataContextConfig(concurrency=ConcurrencyConfig(enabled=True))
    assert data_context_config.concurrency.enabled


def test_data_context_concurrency_property():
    data_context = get_context(
        project_config=DataContextConfig(
            concurrency=ConcurrencyConfig(enabled=True),
            store_backend_defaults=InMemoryStoreBackendDefaults(),
        )
    )
    assert data_context.concurrency.enabled



from __future__ import annotations

import logging
import warnings
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    Union,
)

import great_expectations.exceptions as gx_exceptions
from great_expectations.core._docs_decorators import public_api
from great_expectations.core.id_dict import IDDict
from great_expectations.render import (
    AtomicDiagnosticRendererType,  # noqa: TCH001
    AtomicPrescriptiveRendererType,  # noqa: TCH001
    AtomicRendererType,  # noqa: TCH001
)
from great_expectations.validator.computed_metric import MetricValue  # noqa: TCH001

if TYPE_CHECKING:
    from great_expectations.core import ExpectationConfiguration
    from great_expectations.core.metric_function_types import (
        MetricFunctionTypes,
        MetricPartialFunctionTypes,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation import Expectation
    from great_expectations.expectations.metrics.metric_provider import MetricProvider
    from great_expectations.render import RenderedAtomicContent, RenderedContent

logger = logging.getLogger(__name__)

_registered_expectations: dict = {}
_registered_metrics: dict = {}
_registered_renderers: dict = {}

"""
{
  "metric_name"
    metric_domain_keys
    metric_value_keys
    metric_dependencies
    providers:
      engine: provider
}
"""


class RendererImpl(NamedTuple):
    expectation: str
    renderer: Callable[..., Union[RenderedAtomicContent, RenderedContent]]


def register_renderer(
    object_name: str,
    parent_class: Union[Type[Expectation], Type[MetricProvider]],
    renderer_fn: Callable[..., Union[RenderedAtomicContent, RenderedContent]],
):
    # noinspection PyUnresolvedReferences
    renderer_name = renderer_fn._renderer_type  # type: ignore[attr-defined]
    if object_name not in _registered_renderers:
        logger.debug(f"Registering {renderer_name} for expectation_type {object_name}.")
        _registered_renderers[object_name] = {
            renderer_name: (parent_class, renderer_fn)
        }
        return

    if renderer_name in _registered_renderers[object_name]:
        if _registered_renderers[object_name][renderer_name] == (
            parent_class,
            renderer_fn,
        ):
            logger.info(
                f"Multiple declarations of {renderer_name} renderer for expectation_type {object_name} "
                f"found."
            )
            return
        else:
            logger.warning(
                f"Overwriting declaration of {renderer_name} renderer for expectation_type "
                f"{object_name}."
            )
            _registered_renderers[object_name][renderer_name] = (
                parent_class,
                renderer_fn,
            )
        return
    else:
        logger.debug(f"Registering {renderer_name} for expectation_type {object_name}.")
        _registered_renderers[object_name][renderer_name] = (parent_class, renderer_fn)
        return


def get_renderer_names(expectation_or_metric_type: str) -> List[str]:
    """Gets renderer names for a given Expectation or Metric.

    Args:
        expectation_or_metric_type: The type of an Expectation or Metric for which to get renderer names.

    Returns:
        A list of renderer names for the Expectation or Metric.
    """
    return list(_registered_renderers.get(expectation_or_metric_type, {}).keys())


def get_renderer_names_with_renderer_types(
    expectation_or_metric_type: str,
    renderer_types: List[AtomicRendererType],
) -> List[Union[str, AtomicDiagnosticRendererType, AtomicPrescriptiveRendererType]]:
    """Gets renderer names of a given type, for a given Expectation or Metric.

    Args:
        expectation_or_metric_type: The type of an Expectation or Metric for which to get renderer names.
        renderer_types: The type of the renderers for which to return names.

    Returns:
        A list of renderer names for the given prefixes and Expectation or Metric.
    """
    return [
        renderer_name
        for renderer_name in get_renderer_names(
            expectation_or_metric_type=expectation_or_metric_type
        )
        if any(
            renderer_name.startswith(renderer_type) for renderer_type in renderer_types
        )
    ]


def get_renderer_impls(object_name: str) -> List[str]:
    return list(_registered_renderers.get(object_name, {}).values())


def get_renderer_impl(object_name: str, renderer_type: str) -> Optional[RendererImpl]:
    renderer_tuple: Optional[tuple] = _registered_renderers.get(object_name, {}).get(
        renderer_type
    )
    renderer_impl: Optional[RendererImpl] = None
    if renderer_tuple:
        renderer_impl = RendererImpl(
            expectation=renderer_tuple[0], renderer=renderer_tuple[1]
        )
    return renderer_impl


def register_expectation(expectation: Type[Expectation]) -> None:
    expectation_type = expectation.expectation_type
    # TODO: add version to key
    if expectation_type in _registered_expectations:
        if _registered_expectations[expectation_type] == expectation:
            logger.info(
                f"Multiple declarations of expectation {expectation_type} found."
            )
            return
        else:
            logger.warning(
                f"Overwriting declaration of expectation {expectation_type}."
            )

    logger.debug(f"Registering expectation: {expectation_type}")
    _registered_expectations[expectation_type] = expectation


def _add_response_key(res, key, value):
    if key in res:
        res[key].append(value)
    else:
        res[key] = [value]
    return res


@public_api
def register_metric(
    metric_name: str,
    metric_domain_keys: Tuple[str, ...],
    metric_value_keys: Tuple[str, ...],
    execution_engine: Type[ExecutionEngine],
    metric_class: Type[MetricProvider],
    metric_provider: Optional[Callable],
    metric_fn_type: Optional[
        Union[MetricFunctionTypes, MetricPartialFunctionTypes]
    ] = None,
) -> dict:
    """Register a Metric class for use as a callable metric within Expectations.

    Args:
        metric_name: A name identifying the metric. Metric Name must be globally unique in
            a great_expectations installation.
        metric_domain_keys: A tuple of the keys used to determine the domain of the metric.
        metric_value_keys: A tuple of the keys used to determine the domain of the metric.
        execution_engine: The execution_engine used to execute the metric.
        metric_class: A valid Metric class containing logic to compute attributes of data.
        metric_provider: The MetricProvider class from which the metric_class inherits.
        metric_fn_type: The MetricFunctionType or MetricPartialFunctionType used to define the Metric class.

    Returns:
        A dictionary containing warnings thrown during registration if applicable, and the success status of registration.
    """
    res: dict = {}
    execution_engine_name = execution_engine.__name__
    logger.debug(f"Registering metric: {metric_name}")
    if metric_provider is not None and metric_fn_type is not None:
        metric_provider.metric_fn_type = metric_fn_type  # type: ignore[attr-defined]
    if metric_name in _registered_metrics:
        metric_definition = _registered_metrics[metric_name]
        current_domain_keys = metric_definition.get("metric_domain_keys", set())
        if set(current_domain_keys) != set(metric_domain_keys):
            logger.warning(
                f"metric {metric_name} is being registered with different metric_domain_keys; overwriting metric_domain_keys"
            )
            _add_response_key(
                res,
                "warning",
                f"metric {metric_name} is being registered with different metric_domain_keys; overwriting metric_domain_keys",
            )

        current_value_keys = metric_definition.get("metric_value_keys", set())
        if set(current_value_keys) != set(metric_value_keys):
            logger.warning(
                f"metric {metric_name} is being registered with different metric_value_keys; overwriting metric_value_keys"
            )
            _add_response_key(
                res,
                "warning",
                f"metric {metric_name} is being registered with different metric_value_keys; overwriting metric_value_keys",
            )

        providers = metric_definition.get("providers", {})
        if execution_engine_name in providers:
            current_provider_cls, current_provider_fn = providers[execution_engine_name]
            if current_provider_fn != metric_provider:
                logger.warning(
                    f"metric {metric_name} is being registered with different metric_provider; overwriting metric_provider"
                )
                _add_response_key(
                    res,
                    "warning",
                    f"metric {metric_name} is being registered with different metric_provider; overwriting metric_provider",
                )
                providers[execution_engine_name] = metric_class, metric_provider
            else:
                logger.info(
                    f"Multiple declarations of metric {metric_name} for engine {execution_engine_name}."
                )
                _add_response_key(
                    res,
                    "info",
                    f"Multiple declarations of metric {metric_name} for engine {execution_engine_name}.",
                )
        else:
            providers[execution_engine_name] = metric_class, metric_provider
    else:
        metric_definition = {
            "metric_domain_keys": metric_domain_keys,
            "metric_value_keys": metric_value_keys,
            "default_kwarg_values": metric_class.default_kwarg_values,
            "providers": {execution_engine_name: (metric_class, metric_provider)},
        }
        _registered_metrics[metric_name] = metric_definition

    res["success"] = True

    return res


def get_metric_provider(
    metric_name: str, execution_engine: ExecutionEngine
) -> Tuple[MetricProvider, Callable]:
    try:
        metric_definition = _registered_metrics[metric_name]
        return metric_definition["providers"][type(execution_engine).__name__]
    except KeyError:
        raise gx_exceptions.MetricProviderError(
            f"No provider found for {metric_name} using {type(execution_engine).__name__}"
        )


def get_metric_function_type(
    metric_name: str, execution_engine: ExecutionEngine
) -> Optional[Union[MetricPartialFunctionTypes, MetricFunctionTypes]]:
    try:
        metric_definition = _registered_metrics[metric_name]
        provider_fn, provider_class = metric_definition["providers"][
            type(execution_engine).__name__
        ]
        return getattr(provider_fn, "metric_fn_type", None)
    except KeyError:
        raise gx_exceptions.MetricProviderError(
            f"No provider found for {metric_name} using {type(execution_engine).__name__}"
        )


def get_metric_kwargs(
    metric_name: str,
    configuration: Optional[ExpectationConfiguration] = None,
    runtime_configuration: Optional[dict] = None,
) -> dict:
    try:
        metric_definition = _registered_metrics.get(metric_name)
        if metric_definition is None:
            raise gx_exceptions.MetricProviderError(
                f"No definition found for {metric_name}"
            )
        default_kwarg_values = metric_definition["default_kwarg_values"]
        metric_kwargs = {
            "metric_domain_keys": metric_definition["metric_domain_keys"],
            "metric_value_keys": metric_definition["metric_value_keys"],
        }
        if configuration:
            expectation_impl = get_expectation_impl(configuration.expectation_type)
            configuration_kwargs = expectation_impl().get_runtime_kwargs(
                configuration=configuration, runtime_configuration=runtime_configuration
            )
            if len(metric_kwargs["metric_domain_keys"]) > 0:
                metric_domain_kwargs = IDDict(
                    {
                        k: configuration_kwargs.get(k) or default_kwarg_values.get(k)
                        for k in metric_kwargs["metric_domain_keys"]
                    }
                )
            else:
                metric_domain_kwargs = IDDict()
            if len(metric_kwargs["metric_value_keys"]) > 0:
                metric_value_kwargs = IDDict(
                    {
                        k: configuration_kwargs.get(k)
                        if configuration_kwargs.get(k) is not None
                        else default_kwarg_values.get(k)
                        for k in metric_kwargs["metric_value_keys"]
                    }
                )
            else:
                metric_value_kwargs = IDDict()
            metric_kwargs["metric_domain_kwargs"] = metric_domain_kwargs
            metric_kwargs["metric_value_kwargs"] = metric_value_kwargs
        return metric_kwargs
    except KeyError:
        raise gx_exceptions.MetricProviderError(
            f"Incomplete definition found for {metric_name}"
        )


def get_domain_metrics_dict_by_name(
    metrics: Dict[Tuple[str, str, str], MetricValue], metric_domain_kwargs: IDDict
):
    return {
        metric_edge_key_id_tuple[0]: metric_value
        for metric_edge_key_id_tuple, metric_value in metrics.items()
        if metric_edge_key_id_tuple[1] == metric_domain_kwargs.to_id()
    }


def get_expectation_impl(expectation_name: str) -> Type[Expectation]:
    renamed: Dict[str, str] = {
        "expect_column_values_to_be_vector": "expect_column_values_to_be_vectors",
        "expect_columns_values_confidence_for_data_label_to_be_greater_than_or_equalto_threshold": "expect_column_values_confidence_for_data_label_to_be_greater_than_or_equal_to_threshold",
        "expect_column_values_to_be_greater_than_or_equal_to_threshold": "expect_column_values_to_be_probabilistically_greater_than_or_equal_to_threshold",
    }
    if expectation_name in renamed:
        # deprecated-v0.14.12
        warnings.warn(
            f"Expectation {expectation_name} was renamed to {renamed['expectation_name']} as of v0.14.12 "
            "Please update usage in your pipeline(s) before the v0.17 release",
            DeprecationWarning,
        )
        expectation_name = renamed[expectation_name]

    expectation: Type[Expectation] | None = _registered_expectations.get(
        expectation_name
    )
    if not expectation:
        raise gx_exceptions.ExpectationNotFoundError(f"{expectation_name} not found")

    return expectation


def list_registered_expectation_implementations(
    expectation_root: Optional[Type[Expectation]] = None,
) -> List[str]:
    registered_expectation_implementations = []
    for (
        expectation_name,
        expectation_implementation,
    ) in _registered_expectations.items():
        if expectation_root is None:
            registered_expectation_implementations.append(expectation_name)
        elif expectation_root and issubclass(
            expectation_implementation, expectation_root
        ):
            registered_expectation_implementations.append(expectation_name)

    return registered_expectation_implementations



import difflib
import functools
import logging
import pathlib
from pprint import pformat as pf

import pytest

# apply markers to entire test module
pytestmark = [pytest.mark.integration]

from great_expectations import get_context
from great_expectations.data_context import FileDataContext
from great_expectations.experimental.datasources.config import GxConfig
from great_expectations.experimental.datasources.interfaces import Datasource

logger = logging.getLogger(__file__)


@pytest.fixture
def db_file() -> pathlib.Path:
    relative_path = pathlib.Path(
        "..",
        "..",
        "test_sets",
        "taxi_yellow_tripdata_samples",
        "sqlite",
        "yellow_tripdata.db",
    )
    db_file = pathlib.Path(__file__).parent.joinpath(relative_path).resolve(strict=True)
    assert db_file.exists()
    return db_file


@pytest.fixture
def zep_config_dict(db_file) -> dict:
    return {
        "xdatasources": {
            "my_sql_ds": {
                "connection_string": f"sqlite:///{db_file}",
                "name": "my_sql_ds",
                "type": "sql",
                "assets": {
                    "my_asset": {
                        "name": "my_asset",
                        "table_name": "yellow_tripdata_sample_2019_01",
                        "type": "table",
                        "column_splitter": {
                            "column_name": "pickup_datetime",
                            "method_name": "split_on_year_and_month",
                        },
                        "order_by": [
                            {"key": "year"},
                            {"key": "month"},
                        ],
                    },
                },
            }
        }
    }


@pytest.fixture
def zep_only_config(zep_config_dict: dict) -> GxConfig:
    """Creates a ZEP `GxConfig` object and ensures it contains at least one `Datasource`"""
    zep_config = GxConfig.parse_obj(zep_config_dict)
    assert zep_config.datasources
    return zep_config


@pytest.fixture
def file_dc_config_dir_init(tmp_path: pathlib.Path) -> pathlib.Path:
    """
    Initialize an regular/old-style FileDataContext project config directory.
    Removed on teardown.
    """
    gx_yml = tmp_path / FileDataContext.GX_DIR / FileDataContext.GX_YML
    assert gx_yml.exists() is False
    FileDataContext.create(tmp_path)
    assert gx_yml.exists()

    tmp_gx_dir = gx_yml.parent.absolute()
    logger.info(f"tmp_gx_dir -> {tmp_gx_dir}")
    return tmp_gx_dir


@pytest.fixture
def zep_yaml_config_file(
    file_dc_config_dir_init: pathlib.Path, zep_only_config: GxConfig
) -> pathlib.Path:
    """
    Dump the provided GxConfig to a temporary path. File is removed during test teardown.

    Append ZEP config to default config file
    """
    config_file_path = file_dc_config_dir_init / FileDataContext.GX_YML

    assert config_file_path.exists() is True

    with open(config_file_path, mode="a") as f_append:
        yaml_string = "\n# ZEP\n" + zep_only_config.yaml()
        f_append.write(yaml_string)

    for ds_name in zep_only_config.datasources.keys():
        assert ds_name in yaml_string

    logger.info(f"  Config File Text\n-----------\n{config_file_path.read_text()}")
    return config_file_path


@pytest.fixture
@functools.lru_cache(maxsize=1)
def zep_file_context(zep_yaml_config_file: pathlib.Path) -> FileDataContext:
    context = get_context(
        context_root_dir=zep_yaml_config_file.parent, cloud_mode=False
    )
    assert isinstance(context, FileDataContext)
    return context


def test_load_an_existing_config(
    zep_yaml_config_file: pathlib.Path, zep_only_config: GxConfig
):
    context = get_context(
        context_root_dir=zep_yaml_config_file.parent, cloud_mode=False
    )

    assert context.zep_config == zep_only_config


def test_serialize_zep_config(zep_file_context: FileDataContext):
    dumped_yaml: str = zep_file_context.zep_config.yaml()
    print(f"  Dumped Config\n\n{dumped_yaml}\n")

    assert zep_file_context.zep_config.datasources

    for ds_name, datasource in zep_file_context.zep_config.datasources.items():
        assert ds_name in dumped_yaml

        for asset_name in datasource.assets.keys():
            assert asset_name in dumped_yaml


def test_zep_simple_validate_workflow(zep_file_context: FileDataContext):
    datasource = zep_file_context.get_datasource("my_sql_ds")
    assert isinstance(datasource, Datasource)
    batch_request = datasource.get_asset("my_asset").build_batch_request(
        {"year": 2019, "month": 1}
    )

    validator = zep_file_context.get_validator(batch_request=batch_request)
    result = validator.expect_column_max_to_be_between(
        column="passenger_count", min_value=1, max_value=12
    )
    print(f"  results ->\n{pf(result)}")
    assert result["success"] is True


def test_save_project_does_not_break(zep_file_context: FileDataContext):
    print(zep_file_context.zep_config)
    zep_file_context._save_project_config()


def test_variables_save_config_does_not_break(zep_file_context: FileDataContext):
    print(zep_file_context.zep_config)
    print(zep_file_context.variables)
    zep_file_context.variables.save_config()


def test_save_datacontext_persists_zep_config(
    file_dc_config_dir_init: pathlib.Path, zep_only_config: GxConfig
):
    config_file = file_dc_config_dir_init / FileDataContext.GX_YML

    initial_yaml = config_file.read_text()
    for ds_name in zep_only_config.datasources:
        assert ds_name not in initial_yaml

    context: FileDataContext = get_context(
        context_root_dir=config_file.parent, cloud_mode=False
    )

    context.zep_config = zep_only_config
    context._save_project_config()

    final_yaml = config_file.read_text()
    diff = difflib.ndiff(initial_yaml.splitlines(), final_yaml.splitlines())

    print("\n".join(diff))

    for ds_name in zep_only_config.datasources:
        assert ds_name in final_yaml


def test_add_and_save_zep_datasource(
    file_dc_config_dir_init: pathlib.Path,
    zep_only_config: GxConfig,
    db_file: pathlib.Path,
):
    datasource_name = "save_ds_test"
    config_file = file_dc_config_dir_init / FileDataContext.GX_YML

    initial_yaml = config_file.read_text()
    assert datasource_name not in initial_yaml

    context: FileDataContext = get_context(
        context_root_dir=config_file.parent, cloud_mode=False
    )

    ds = context.sources.add_sqlite(
        name=datasource_name, connection_string=f"sqlite:///{db_file}"
    )
    context._save_project_config()

    final_yaml = config_file.read_text()
    diff = difflib.ndiff(initial_yaml.splitlines(), final_yaml.splitlines())

    print("\n".join(diff))

    assert datasource_name == ds.name
    assert datasource_name in final_yaml
    # ensure comments preserved
    assert "# Welcome to Great Expectations!" in final_yaml


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



import logging
from typing import List
from unittest import mock

import pandas as pd
import pytest

import great_expectations.exceptions as gx_exceptions
from great_expectations import DataContext
from great_expectations.core import (
    ExpectationConfiguration,
    ExpectationSuite,
    ExpectationSuiteValidationResult,
    ExpectationValidationResult,
)
from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.core.usage_statistics.usage_statistics import (
    UsageStatisticsHandler,
)
from great_expectations.validator.validator import Validator

logger = logging.getLogger(__name__)

try:
    from pyspark.sql import DataFrame
except ImportError:
    DataFrame = None

    logger.debug(
        "Unable to load pyspark; install optional spark dependency for support."
    )


@pytest.fixture
def test_pandas_df():
    df: pd.DataFrame = pd.DataFrame(
        data=[["Scott"], ["Jeff"], ["Thomas"], ["Ann"]], columns=["Name"]
    )
    return df


@pytest.fixture
def test_spark_df(test_pandas_df, spark_session):
    df: DataFrame = spark_session.createDataFrame(data=test_pandas_df)
    return df


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_catch_exceptions_no_exceptions(
    mock_emit, in_memory_runtime_context, test_spark_df
):
    catch_exceptions: bool = False  # expect exceptions to be raised
    result_format: dict = {
        "result_format": "SUMMARY",
    }
    runtime_environment_arguments = {
        "catch_exceptions": catch_exceptions,
        "result_format": result_format,
    }

    suite: ExpectationSuite = in_memory_runtime_context.add_expectation_suite(
        "test_suite"
    )

    expectation_configuration: ExpectationConfiguration

    expectation_meta: dict = {"Notes": "Some notes"}

    expectation_arguments_without_meta: dict

    expectation_arguments_column: dict = {
        "include_config": True,
        "column": "Name",  # use correct column to avoid error
    }
    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_column_values_to_not_be_null",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    expectation_arguments_table: dict = {
        "include_config": True,
        "value": 4,
    }
    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_table
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_table_row_count_to_equal",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    runtime_batch_request = RuntimeBatchRequest(
        datasource_name="spark_datasource",
        data_connector_name="runtime_data_connector",
        data_asset_name="insert_your_data_asset_name_here",
        runtime_parameters={"batch_data": test_spark_df},
        batch_identifiers={
            "id_key_0": "id_value_0",
            "id_key_1": "id_value_1",
        },
    )

    validator: Validator = in_memory_runtime_context.get_validator(
        batch_request=runtime_batch_request,
        expectation_suite=suite,
    )

    # Test calling "validator.validate()" explicitly.

    validator_validation: ExpectationSuiteValidationResult = validator.validate(
        **runtime_environment_arguments
    )
    results: List[ExpectationValidationResult] = validator_validation.results
    assert len(results) == 2

    result: ExpectationValidationResult

    for result in results:
        assert result.success
        assert (
            "exception_traceback" not in result.exception_info
        ) or not result.exception_info["exception_traceback"]
        assert (
            "exception_message" not in result.exception_info
        ) or not result.exception_info["exception_message"]

    # Test calling "validator.expect_*" through "validator.validate_expectation()".

    expectation_parameters: dict

    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_parameters = dict(
        **expectation_arguments_without_meta, **expectation_meta
    )
    result = validator.expect_column_values_to_not_be_null(**expectation_parameters)
    assert result.success

    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_table
    )
    expectation_parameters = dict(
        **expectation_arguments_without_meta, **expectation_meta
    )
    result = validator.expect_table_row_count_to_equal(**expectation_parameters)
    assert result.success

    # In-Memory DataContext does not have UsageStatisticsHandler configured
    assert mock_emit.call_count == 0


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_catch_exceptions_exception_occurred_catch_exceptions_false(
    mock_emit, in_memory_runtime_context, test_spark_df
):
    catch_exceptions: bool = False  # expect exceptions to be raised
    result_format: dict = {
        "result_format": "SUMMARY",
    }
    runtime_environment_arguments = {
        "catch_exceptions": catch_exceptions,
        "result_format": result_format,
    }

    suite: ExpectationSuite = in_memory_runtime_context.add_expectation_suite(
        "test_suite"
    )

    expectation_configuration: ExpectationConfiguration

    expectation_meta: dict = {"Notes": "Some notes"}

    expectation_arguments_without_meta: dict

    expectation_arguments_column: dict = {
        "include_config": True,
        "column": "unknown_column",  # use intentionally incorrect column to force error in "MetricProvider" evaluations
    }
    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_column_values_to_not_be_null",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    expectation_arguments_table: dict = {
        "include_config": True,
        "value": 4,
    }
    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_table
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_table_row_count_to_equal",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    runtime_batch_request = RuntimeBatchRequest(
        datasource_name="spark_datasource",
        data_connector_name="runtime_data_connector",
        data_asset_name="insert_your_data_asset_name_here",
        runtime_parameters={"batch_data": test_spark_df},
        batch_identifiers={
            "id_key_0": "id_value_0",
            "id_key_1": "id_value_1",
        },
    )

    validator: Validator = in_memory_runtime_context.get_validator(
        batch_request=runtime_batch_request,
        expectation_suite=suite,
    )

    expected_exception_message: str = (
        'Error: The column "unknown_column" in BatchData does not exist.'
    )

    # Test calling "validator.validate()" explicitly.

    with pytest.raises(gx_exceptions.MetricResolutionError) as e:
        # noinspection PyUnusedLocal
        validator_validation: ExpectationSuiteValidationResult = validator.validate(
            **runtime_environment_arguments
        )
    assert e.value.message == expected_exception_message

    # Test calling "validator.expect_*" through "validator.validate_expectation()".

    expectation_parameters: dict

    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_parameters = dict(
        **expectation_arguments_without_meta, **expectation_meta
    )

    with pytest.raises(gx_exceptions.MetricResolutionError) as e:
        # noinspection PyUnusedLocal
        result: ExpectationValidationResult = (
            validator.expect_column_values_to_not_be_null(**expectation_parameters)
        )
    assert e.value.message == expected_exception_message

    # Confirm that even though exceptions may occur in some expectations, other expectations can be validated properly.

    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_table
    )
    expectation_parameters = dict(
        **expectation_arguments_without_meta, **expectation_meta
    )
    result: ExpectationValidationResult = validator.expect_table_row_count_to_equal(
        **expectation_parameters
    )
    assert result.success

    # In-Memory DataContext does not have UsageStatisticsHandler configured
    assert mock_emit.call_count == 0


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_catch_exceptions_exception_occurred_catch_exceptions_true(
    mock_emit, in_memory_runtime_context, test_spark_df
):
    catch_exceptions: bool = True  # expect exceptions to be caught
    result_format: dict = {
        "result_format": "SUMMARY",
    }
    runtime_environment_arguments = {
        "catch_exceptions": catch_exceptions,
        "result_format": result_format,
    }

    suite: ExpectationSuite = in_memory_runtime_context.add_expectation_suite(
        "test_suite"
    )

    expectation_configuration: ExpectationConfiguration

    expectation_meta: dict = {"Notes": "Some notes"}

    expectation_arguments_without_meta: dict

    expectation_arguments_column: dict = {
        "include_config": True,
        "column": "unknown_column",  # use intentionally incorrect column to force error in "MetricProvider" evaluations
    }
    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_column_values_to_not_be_null",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    expectation_arguments_table: dict = {
        "include_config": True,
        "value": 4,
    }
    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_table
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_table_row_count_to_equal",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    runtime_batch_request = RuntimeBatchRequest(
        datasource_name="spark_datasource",
        data_connector_name="runtime_data_connector",
        data_asset_name="insert_your_data_asset_name_here",
        runtime_parameters={"batch_data": test_spark_df},
        batch_identifiers={
            "id_key_0": "id_value_0",
            "id_key_1": "id_value_1",
        },
    )

    validator: Validator = in_memory_runtime_context.get_validator(
        batch_request=runtime_batch_request,
        expectation_suite=suite,
    )

    expected_exception_message: str = (
        'Error: The column "unknown_column" in BatchData does not exist.'
    )

    # Test calling "validator.validate()" explicitly.

    validator_validation: ExpectationSuiteValidationResult = validator.validate(
        **runtime_environment_arguments
    )
    results: List[ExpectationValidationResult] = validator_validation.results
    assert len(results) == 2

    # Confirm that even though an exception occurred in one expectation, the other expectation is validated properly.

    results = sorted(
        results, key=lambda element: element.expectation_config["expectation_type"]
    )

    result: ExpectationValidationResult

    result = results[0]
    assert (
        result.expectation_config["expectation_type"]
        == "expect_column_values_to_not_be_null"
    )
    assert not result.success
    assert "exception_traceback" in result.exception_info
    assert "exception_message" in result.exception_info
    assert result.exception_info["exception_message"] == expected_exception_message

    result = results[1]
    assert (
        result.expectation_config["expectation_type"]
        == "expect_table_row_count_to_equal"
    )
    assert result.success
    assert (
        "exception_traceback" not in result.exception_info
    ) or not result.exception_info["exception_traceback"]
    assert (
        "exception_message" not in result.exception_info
    ) or not result.exception_info["exception_message"]

    # Test calling "validator.expect_*" through "validator.validate_expectation()".

    expectation_parameters: dict

    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_parameters = dict(
        **expectation_arguments_without_meta, **expectation_meta
    )
    result = validator.expect_column_values_to_not_be_null(**expectation_parameters)
    assert not result.success
    assert "exception_traceback" in result.exception_info
    assert "exception_message" in result.exception_info
    assert result.exception_info["exception_message"] == expected_exception_message

    # Confirm that even though exceptions may occur in some expectations, other expectations can be validated properly.

    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_table
    )
    expectation_parameters = dict(
        **expectation_arguments_without_meta, **expectation_meta
    )
    result = validator.expect_table_row_count_to_equal(**expectation_parameters)
    assert result.success
    assert (
        "exception_traceback" not in result.exception_info
    ) or not result.exception_info["exception_traceback"]
    assert (
        "exception_message" not in result.exception_info
    ) or not result.exception_info["exception_message"]

    # In-Memory DataContext does not have UsageStatisticsHandler configured
    assert mock_emit.call_count == 0


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_result_format_configured_no_set_default_override(
    mock_emit, in_memory_runtime_context, test_spark_df
):
    catch_exceptions: bool = False  # expect exceptions to be raised
    result_format: dict

    result_format = {
        "result_format": "SUMMARY",
    }
    runtime_environment_arguments: dict = {
        "catch_exceptions": catch_exceptions,
        "result_format": result_format,
    }

    suite: ExpectationSuite

    suite = in_memory_runtime_context.add_expectation_suite("test_suite")

    expectation_configuration: ExpectationConfiguration

    expectation_meta: dict = {"Notes": "Some notes"}

    expectation_arguments_without_meta: dict

    expectation_arguments_column: dict = {
        "include_config": True,
        "column": "Name",  # use correct column to avoid error
    }
    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_column_values_to_not_be_null",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    runtime_batch_request = RuntimeBatchRequest(
        datasource_name="spark_datasource",
        data_connector_name="runtime_data_connector",
        data_asset_name="insert_your_data_asset_name_here",
        runtime_parameters={"batch_data": test_spark_df},
        batch_identifiers={
            "id_key_0": "id_value_0",
            "id_key_1": "id_value_1",
        },
    )

    validator: Validator

    validator = in_memory_runtime_context.get_validator(
        batch_request=runtime_batch_request,
        expectation_suite=suite,
    )

    # Test calling "validator.validate()" explicitly.

    validator_validation: ExpectationSuiteValidationResult

    validator_validation = validator.validate(**runtime_environment_arguments)

    results: List[ExpectationValidationResult]

    results = validator_validation.results
    assert len(results) == 1

    result: ExpectationValidationResult

    result = results[0]
    assert result.success
    assert len(result.result.keys()) > 0
    assert result.result == {
        "element_count": 4,
        "unexpected_count": 0,
        "unexpected_percent": 0.0,
        "partial_unexpected_list": [],
        "partial_unexpected_counts": [],
    }

    result_format = {
        "result_format": "BASIC",
    }
    runtime_environment_arguments: dict = {
        "catch_exceptions": catch_exceptions,
        "result_format": result_format,
    }

    suite = in_memory_runtime_context.add_or_update_expectation_suite("test_suite")

    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_column_values_to_not_be_null",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    validator = in_memory_runtime_context.get_validator(
        batch_request=runtime_batch_request,
        expectation_suite=suite,
    )

    validator_validation = validator.validate(**runtime_environment_arguments)

    results = validator_validation.results
    assert len(results) == 1

    result = results[0]
    assert result.success
    assert len(result.result.keys()) > 0
    assert result.result == {
        "element_count": 4,
        "unexpected_count": 0,
        "unexpected_percent": 0.0,
        "partial_unexpected_list": [],
    }

    result_format = {
        "result_format": "BOOLEAN_ONLY",
    }
    runtime_environment_arguments: dict = {
        "catch_exceptions": catch_exceptions,
        "result_format": result_format,
    }

    suite = in_memory_runtime_context.add_or_update_expectation_suite("test_suite")

    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_column_values_to_not_be_null",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )

    suite.add_expectation(expectation_configuration=expectation_configuration)

    validator = in_memory_runtime_context.get_validator(
        batch_request=runtime_batch_request,
        expectation_suite=suite,
    )

    validator_validation = validator.validate(**runtime_environment_arguments)

    results = validator_validation.results
    assert len(results) == 1

    result = results[0]
    assert result.success
    assert result.to_json_dict() == {
        "expectation_config": {
            "kwargs": {
                "catch_exceptions": False,
                "result_format": {"result_format": "BOOLEAN_ONLY"},
                "include_config": True,
                "column": "Name",
                "batch_id": "bd7b9290f981fde37aabd403e8a507ea",
            },
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {"Notes": "Some notes"},
        },
        "meta": {},
        "exception_info": {
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
        "result": {},
        "success": True,
    }
    assert len(result.result.keys()) == 0
    assert result.result == {}

    # Test calling "validator.expect_*" through "validator.validate_expectation()".

    expectation_parameters: dict

    expectation_parameters = dict(
        **expectation_arguments_without_meta, **expectation_meta
    )
    result = validator.expect_column_values_to_not_be_null(**expectation_parameters)
    assert result.success
    assert result.to_json_dict() == {
        "success": True,
        "meta": {},
        "expectation_config": {
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {},
            "kwargs": {
                "catch_exceptions": False,
                "result_format": {
                    "result_format": "BOOLEAN_ONLY",
                    "include_unexpected_rows": False,
                    "partial_unexpected_count": 20,
                },
                "include_config": True,
                "column": "Name",
                "Notes": "Some notes",
                "batch_id": "bd7b9290f981fde37aabd403e8a507ea",
            },
        },
        "result": {},
        "exception_info": {
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
    }
    assert len(result.result.keys()) == 0
    assert result.result == {}

    # In-Memory DataContext does not have UsageStatisticsHandler configured
    assert mock_emit.call_count == 0


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_result_format_configured_with_set_default_override(
    mock_emit, in_memory_runtime_context, test_spark_df
):
    catch_exceptions: bool = False  # expect exceptions to be raised
    result_format: dict

    result_format = {
        "result_format": "SUMMARY",
    }
    runtime_environment_arguments: dict = {
        "catch_exceptions": catch_exceptions,
        "result_format": result_format,
    }

    suite: ExpectationSuite

    suite = in_memory_runtime_context.add_expectation_suite("test_suite")

    expectation_configuration: ExpectationConfiguration

    expectation_meta: dict = {"Notes": "Some notes"}

    expectation_arguments_without_meta: dict

    expectation_arguments_column: dict = {
        "include_config": True,
        "column": "Name",  # use correct column to avoid error
    }
    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_column_values_to_not_be_null",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    runtime_batch_request = RuntimeBatchRequest(
        datasource_name="spark_datasource",
        data_connector_name="runtime_data_connector",
        data_asset_name="insert_your_data_asset_name_here",
        runtime_parameters={"batch_data": test_spark_df},
        batch_identifiers={
            "id_key_0": "id_value_0",
            "id_key_1": "id_value_1",
        },
    )

    validator: Validator

    validator = in_memory_runtime_context.get_validator(
        batch_request=runtime_batch_request,
        expectation_suite=suite,
    )

    validator.set_default_expectation_argument("result_format", "BOOLEAN_ONLY")

    # Test calling "validator.validate()" explicitly.

    validator_validation: ExpectationSuiteValidationResult

    validator_validation = validator.validate()

    results: List[ExpectationValidationResult]

    results = validator_validation.results
    assert len(results) == 1

    result: ExpectationValidationResult

    result = results[0]
    assert result.success
    assert result.to_json_dict() == {
        "result": {},
        "expectation_config": {
            "kwargs": {
                "catch_exceptions": False,
                "result_format": {"result_format": "SUMMARY"},
                "include_config": True,
                "column": "Name",
                "batch_id": "bd7b9290f981fde37aabd403e8a507ea",
            },
            "meta": {"Notes": "Some notes"},
            "expectation_type": "expect_column_values_to_not_be_null",
        },
        "success": True,
        "meta": {},
        "exception_info": {
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
    }
    assert len(result.result.keys()) == 0
    assert result.result == {}

    result_format = {
        "result_format": "BASIC",
    }
    runtime_environment_arguments: dict = {
        "catch_exceptions": catch_exceptions,
        "result_format": result_format,
    }

    suite = in_memory_runtime_context.add_or_update_expectation_suite("test_suite")

    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_column_values_to_not_be_null",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    validator = in_memory_runtime_context.get_validator(
        batch_request=runtime_batch_request,
        expectation_suite=suite,
    )

    validator.set_default_expectation_argument("result_format", "BOOLEAN_ONLY")

    validator_validation = validator.validate()

    results = validator_validation.results
    assert len(results) == 1

    result = results[0]
    assert len(result.result.keys()) == 0

    # Test calling "validator.expect_*" through "validator.validate_expectation()".

    expectation_parameters: dict

    expectation_parameters = dict(**expectation_arguments_column, **expectation_meta)
    result = validator.expect_column_values_to_not_be_null(**expectation_parameters)
    assert result.success
    assert result.to_json_dict() == {
        "result": {},
        "expectation_config": {
            "kwargs": {
                "include_config": True,
                "column": "Name",
                "Notes": "Some notes",
                "batch_id": "bd7b9290f981fde37aabd403e8a507ea",
            },
            "meta": {},
            "expectation_type": "expect_column_values_to_not_be_null",
        },
        "success": True,
        "meta": {},
        "exception_info": {
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
    }
    assert len(result.result.keys()) == 0
    assert result.result == {}

    # In-Memory DataContext does not have UsageStatisticsHandler configured
    assert mock_emit.call_count == 0


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_in_memory_runtime_context_configured_with_usage_stats_handler(
    mock_emit, in_memory_runtime_context, test_pandas_df
):
    context: DataContext = in_memory_runtime_context

    # manually set usage statistics handler
    handler = UsageStatisticsHandler(
        data_context=context,
        data_context_id=context._data_context_id,
        usage_statistics_url="http://fakeendpoint.com",
    )
    context._usage_statistics_handler = handler

    catch_exceptions: bool = False  # expect exceptions to be raised
    result_format: dict = {
        "result_format": "SUMMARY",
    }
    runtime_environment_arguments = {
        "catch_exceptions": catch_exceptions,
        "result_format": result_format,
    }

    suite: ExpectationSuite = in_memory_runtime_context.add_expectation_suite(
        "test_suite"
    )

    expectation_configuration: ExpectationConfiguration

    expectation_meta: dict = {"Notes": "Some notes"}

    expectation_arguments_without_meta: dict

    expectation_arguments_column: dict = {
        "include_config": True,
        "column": "Name",  # use correct column to avoid error
    }
    expectation_arguments_without_meta = dict(
        **runtime_environment_arguments, **expectation_arguments_column
    )
    expectation_configuration = ExpectationConfiguration(
        expectation_type="expect_column_values_to_not_be_null",
        kwargs=expectation_arguments_without_meta,
        meta=expectation_meta,
    )
    suite.add_expectation(expectation_configuration=expectation_configuration)

    # emit 1 from add_expectation
    assert mock_emit.call_count == 1
    assert mock_emit.call_args_list == [
        mock.call(
            {
                "event": "expectation_suite.add_expectation",
                "event_payload": {},
                "success": True,
            }
        )
    ]



import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import Union

from great_expectations.data_context.types.base import BaseYamlConfig  # noqa: TCH001
from great_expectations.util import filter_properties_dict

logger = logging.getLogger(__name__)


class ConfigOutputModes(str, Enum):
    TYPED = "typed"
    COMMENTED_MAP = "commented_map"
    YAML = "yaml"
    DICT = "dict"
    JSON_DICT = "json_dict"


class ConfigPeer(ABC):
    """
    A ConfigPeer is an object, whose subclasses can be instantiated using instantiate_class_from_config() (located in
    great_expectations/util.py).  Its immediate descendant subclass must use a subclass of BaseYamlConfig as an argument
    to its constructor, and the subsequent descendants must use only primitive types as their constructor arguments,
    wherever keys correspond to the keys of the "BaseYamlConfig" configuration object counterpart. The name ConfigPeer
    means: Every immediate descendant subclass must have Marshmallow Schema validated configuration class as its peer.

    # TODO: <Alex>2/11/2022</Alex>
    When -- as part of a potential future architecture update -- serialization is decoupled from configuration, the
    configuration objects, persistable as YAML files, will no longer inherit from the BaseYamlConfig class.  Rather,
    any form of serialization (YAML, JSON, SQL Database Tables, Pickle, etc.) will apply as peers, independent of the
    configuration classes themselves.  Hence, as part of this change, ConfigPeer will cease being the superclass of
    business objects (such as BaseDataContext, BaseCheckpoint, and BaseRuleBasedProfiler).  Instead, every persistable
    business object will contain a reference to its corresponding peer class, supporting the ConfigPeer interfaces.
    """

    @property
    @abstractmethod
    def config(self) -> BaseYamlConfig:
        pass

    def get_config(
        self,
        mode: ConfigOutputModes = ConfigOutputModes.TYPED,
        **kwargs,
    ) -> Union[BaseYamlConfig, dict, str]:
        if isinstance(mode, str):
            mode = ConfigOutputModes(mode.lower())

        config: BaseYamlConfig = self.config

        if mode == ConfigOutputModes.TYPED:
            return config

        if mode == ConfigOutputModes.COMMENTED_MAP:
            return config.commented_map

        if mode == ConfigOutputModes.YAML:
            return config.to_yaml_str()

        if mode == ConfigOutputModes.DICT:
            config_kwargs: dict = config.to_dict()
        elif mode == ConfigOutputModes.JSON_DICT:
            config_kwargs: dict = config.to_json_dict()  # type: ignore[no-redef]
        else:
            raise ValueError(f'Unknown mode {mode} in "BaseCheckpoint.get_config()".')

        kwargs["inplace"] = True
        filter_properties_dict(
            properties=config_kwargs,
            **kwargs,
        )

        return config_kwargs

    def __repr__(self) -> str:
        return str(self.get_config())
